﻿// find isometric calculations here: http://clintbellanger.net/articles/isometric_math/
// potentially look at this sometime: http://jsfiddle.net/loktar/63QZz/
/* global webAppCore */
/* global ParticleEngine2 */
(function(){
    /** @global */
    window.Scroll2d = new Scroll2d();
    
    var canvas;
    var context;

    var staticCanvas;
    var staticContext;
    var staticTilemap;

    var appScale = 1;
    
    var fullCanvas;
    var fullContext;
    var fullMapRender = false;
    var fullRenderCallback = null;

    var usingTouch = false;

    var fpslimiter = 1;
    var fpscounter = 0;

    var π = Math.PI;
    var twoPi = π * 2;

    var usedRenderObjects = [];

    var zoomLevel = 1;
    
    var viewPadding = 1;

    var gridSize = 40;
    
    var viewX = 0;
    var viewY = 0;
    
    var updateTime = 16;

    var mapWidth = 10;
    var mapHeight = 10;
    
    var mapTotalWidth = 0;
    var mapTotalHeight = 0;
    
    var xMin = 0;
    var yMin = 0;
    var xLimit = 0;
    var yLimit = 0;
    
    var drawFunction;
    var clickFunction;
    var rightClickFunction;
    var hoverFunction;
    var holdFunction;
    var paintFunction;
    var updateFunction;
    var selectionFunction;

    var particleEngine;

    var lightCanvas;
    var lightContext;

    var sceneDidChange = false;
    
    var startX;
    var startY;
    var curX;
    var curY;
    
    var hx;
    var hy;
    var difX;
    var difY;

    var preRenderCanvas = null;
    var preRenderContext = null;

    var lightFinalRendererCanvas = null;
    var lightFinalRendererContext = null;

    var gradientFilterCanvas = null;
    var gradientFilterContext = null;

    var staticMaxZIndex = 1;


    var selectionX;
    var selectionY;

    var selectionTopX;
    var selectionTopY;
    var selectionWidth;
    var selectionHeight;

    var selectionProgressFunction;

    var selectionColor = "#4CAF50";
    
    var isSelection = false;
    var isPainting = false;
    var didMove = false;
    var isDown = false;
    var isHold = false;

    var isometricMode = false;
    
    var filterOverlay;

    var useFastRenderMode = false;
    
    
    var hdMode = false;
    
    var textPoppers;
    var usedPopperObjects = [];
    
    var holdTimer;

    var doubleHeightTiles = false;
    
    var winW;
    var winH;

    var canvasHolder;
    var usingHolder = false;
    var fitGridToHeight = false;

    var currentViewBounds = null;
    
    /**
     * Creates a new Scroll2d instance.
     * @class
     */
    function Scroll2d() {
        
        this.init = initApp;
        
        this.setMapDims = setMapDims;
        this.drawStaticTile = drawStaticTile;
        this.drawTile = drawTile;
        this.setDrawFunction = setDrawFunction;
        this.setClickFunction = setClickFunction;
        this.setRightClickFunction = setRightClickFunction;
        this.centerOnCoord = centerOnCoord;
        this.checkVisibility = checkVisibility;
        this.setViewPadding = setViewPadding;
        this.setHoverFunction = setHoverFunction;
        this.setHoldFunction = setHoldFunction;
        this.setGridSize = setGridSize;
        this.setDoubleHeightTiles = setDoubleHeightTiles;
        this.drawSprite = drawSprite;
        this.alertPainting = alertPainting;
        this.setPaintFunction = setPaintFunction;
        this.setHDMode = setHDMode;
        this.popText = popText;
        this.setFixedScale = setFixedScale;
        this.drawSquare = drawSquare;
        this.setSnowing = setSnowing;
        this.setEmbers = setEmbers;
        this.renderFullMap = renderFullMap;
        this.setIsometric = setIsometric;
        this.getFPS = getFPS;
        this.setUpdateFunction = setUpdateFunction;
        this.setFitGridToHeight = setFitGridToHeight;
        this.setColorFilter = setColorFilter;
        this.forceSceneChange = forceSceneChange;
        this.setSelectionFunction = setSelectionFunction;
        this.clearOutPaint = clearOutPaint;
        this.initParticleEngine = initParticleEngine;
        this.playParticleEffect = playParticleEffect;
        this.doZoom = doZoom;
        this.sizeElements = sizeElements;
        this.drawLine = drawLine;
        this.setBorderColor = setBorderColor;

        this.setFastRenderMode = setFastRenderMode;

        this.Chevron = Chevron;

        this.getViewBounds = getViewBounds;

        this.getZoomLevel = getZoomLevel;

        this.startPan = startPan;
        this.endPan = endPan;
        
        this.takeScreenshot = takeScreenshot;

        this.setZoomLevel = setZoomLevel;
        this.setFPSLimiter = setFPSLimiter;

        this.modifyViewX = modifyViewX;
        this.modifyViewY = modifyViewY;

        this.getUsingTouch = getUsingTouch;

        this.gamepadDown = gamepadDown;
        this.gamepadUp = gamepadUp;

        this.drawLight = drawLight;
    }

    /**
     * Initialize the Scroll2d instance
     * @param {object} [holder] - A DOM element (DIV generally) that will hold the render canvas.  If not specified, engine looks for a canvas with id "canvas"
     */
    function initApp(holder) {

        particleEngine = ParticleEngine2.getInstance();

        staticMaxZIndex = 1;

        if(window.devicePixelRatio) {
            appScale = window.devicePixelRatio;
        }

        if(typeof webAppCore !== "undefined") {
            if(webAppCore.getPlatform().platform == "desktop" || webAppCore.getPlatform().platform == "web" || webAppCore.getPlatform().platform == "windows") {
                if (window.PointerEvent) {
                    window.addEventListener("pointermove",desktopWindowPointerMove,false);
                } else {
                    window.addEventListener("mousemove",desktopWindowPointerMove,false);
                }
            }
        }

        usedRenderObjects = [];
        
        window.addEventListener("resize",sizeElements,false);

        filterOverlay = document.createElement("div");
        filterOverlay.style.position = "absolute";
        normalizeFilerOverlaySize();

        lightCanvas = document.createElement("canvas");
        lightContext = lightCanvas.getContext("2d");

        gradientFilterCanvas = document.createElement("canvas");
        gradientFilterContext = gradientFilterCanvas.getContext("2d");

        preRenderCanvas = document.createElement("canvas");
        preRenderContext = preRenderCanvas.getContext("2d");

        lightFinalRendererCanvas = document.createElement("canvas");
        lightFinalRendererContext = lightFinalRendererCanvas.getContext("2d");

        if(holder) {
            usingHolder = true;
            canvasHolder = holder;

            staticCanvas = document.createElement("canvas");
            staticContext = staticCanvas.getContext("2d");

            canvas = document.createElement("canvas");
            canvasHolder.appendChild(canvas);
            canvasHolder.appendChild(filterOverlay);
        } else {
            canvas = document.getElementById("canvas");
            
            try{
                canvas.parentNode.insertBefore(filterOverlay, canvas.nextSibling);
            } catch(err) {
                document.appendChild(filterOverlay);
            }
        }
        
        canvas.style.position = "absolute";
        canvas.style.top = "0px";
        canvas.style.left = "0px";
        canvas.style.height = "100%";
        canvas.style.width = "100%";

        if(usingHolder) {
            staticCanvas.style.position = "absolute";
            staticCanvas.style.top = "0px";
            staticCanvas.style.left = "0px";
            staticCanvas.style.height = "100%";
            staticCanvas.style.width = "100%";
        }
        
        context = canvas.getContext("2d");
        
        if(typeof webAppCore !== "undefined") {
            webAppCore.handleInput(filterOverlay,down,move,up,null);
        } else {
            handleInputNonWac(filterOverlay,down,move,up,null);
        }
    
        filterOverlay.addEventListener("wheel",onMouseWheel,false);
        filterOverlay.addEventListener("contextmenu",function(e) {
            e.preventDefault();
        },false);

        sizeElements();
        checkBounds();
        window.requestAnimationFrame(render);
        
        textPoppers = [];

    }

    function initParticleEngine() {
        usingParticleEnging = true;
    }

    function setFPSLimiter(value) {
        fpslimiter = value;
    }

    function setFastRenderMode(mode) {
        useFastRenderMode = mode;
    }

    function setBorderColor(color) {
        if(!color || color == null) {
            filterOverlay.style.border = "none";
            normalizeFilerOverlaySize();
        } else {
            filterOverlay.style.top = "0px";
            filterOverlay.style.left = "0px";
            filterOverlay.style.height = "calc(100% - 8px)";
            filterOverlay.style.width = "calc(100% - 8px)";

            filterOverlay.style.border = "4px solid " + color;
        }
    }

    function normalizeFilerOverlaySize() {
        filterOverlay.style.top = "0px";
        filterOverlay.style.left = "0px";
        filterOverlay.style.height = "100%";
        filterOverlay.style.width = "100%";
    }

    var filterColor = null;
    var filterAmount = 0;

    function setColorFilter(filter,strength) {

        darknessMaskColor = null;

        if(!filter || filter == null) {
            filterColor = null;
        } else {
            if(strength) {
                filterAmount = strength;
            } else {
                filterAmount = 0.1;
            }

            filterColor = filter;
        }
    }
    
    function hexToRgb(hex) {

        var result = /^#?([a-f\d]{2})([a-f\d]{2})([a-f\d]{2})$/i.exec(hex);
        return result ? {
            r: parseInt(result[1], 16),
            g: parseInt(result[2], 16),
            b: parseInt(result[3], 16)
        } : null;
    }

    var lightSources = [];

    function drawLight(x,y,radius,color,tX,tY,step,intensity) {

        if(!color) {
            color = "#ffffff";
        }

        if(!intensity) {
            intensity = 0.2;
        }

        var rgbcolor = hexToRgb(color);

        if(!tX) {
            tX = x;
            tY = y;
        }

        if(!radius) {
            radius = relativeGridSize;
        } else {
            radius = radius * relativeGridSize;
        }

        var ux = x;
        var uy = y;

        if(isometricMode) {
            ux = (x - y) * halfrgs;
            uy = (x + y) * (relativeGridSize / 4);
        } else {
            ux = x * relativeGridSize;
            uy = ((y * relativeGridSize) + relativeGridSize) - halfrgs;
        }

        if(x != tX || y != tY) {
            var tenthGrid = relativeGridSize / 20;
            var stepDistance = tenthGrid * step;

            if(isometricMode) {
                if(tX > x) {
                    ux += stepDistance / 2;
                    uy += stepDistance / 4;
                }
                
                if(tX < x) {
                    ux -= stepDistance / 2;
                    uy -= stepDistance / 4;
                }
                
                if(tY > y) {
                    ux -= stepDistance / 2;
                    uy += stepDistance / 4;
                }
                
                if(tY < y) {
                    ux += stepDistance / 2;
                    uy -= stepDistance / 4;
                }

            } else {
                if(tX > x) {
                    ux += stepDistance;
                }
                
                if(tX < x) {
                    ux -= stepDistance;
                }
                
                if(tY > y) {
                    uy += stepDistance;
                }
                
                if(tY < y) {
                    uy -= stepDistance;
                }
            }
            
        }

        lightSources.push(getFreshLightInstruction(ux,uy,radius,rgbcolor,intensity));

    }

    function getFreshLightInstruction(x,y,r,c,i) {
        if(lightInstructionRecycling.length > 0) {
            var fresh = lightInstructionRecycling.shift();
            fresh.x = x;
            fresh.y = y;
            fresh.radius = r;
            fresh.color = c;
            fresh.intensity = i;
            return fresh;
        } else {
            return new LightInstruction(x,y,r,c,i);
        }
    }

    var lightInstructionRecycling = [];

    function LightInstruction(x,y,r,c,i) {
        this.x = x;
        this.y = y;
        this.radius = r;
        this.color = c;
        this.intensity = i;
    }

    var darknessMaskColor = null;

    function ligthenGradient(ctx, x, y, radius, from, off) {
    
        var radialGradient;

        if(isometricMode) {
            if(radius > (relativeGridSize * 0.8)) {
                ctx.save();

                ctx.transform(1,0,0,0.5,0,0);

                radialGradient = ctx.createRadialGradient(x, y * 2, 0, x, y * 2, radius);
    
                radialGradient.addColorStop(0.0, from);
                radialGradient.addColorStop(1,off);
        
                ctx.fillStyle = radialGradient;
                ctx.beginPath();
                
                ctx.arc(x, y * 2, radius, 0, twoPi);

                ctx.fill();
                ctx.restore();

            } else {

                radialGradient = ctx.createRadialGradient(x, y, 0, x, y, radius);
   
                radialGradient.addColorStop(0.0, from);
                radialGradient.addColorStop(1,off);
        
                ctx.fillStyle = radialGradient;
                ctx.beginPath();

                ctx.arc(x, y, radius, 0, twoPi);
                ctx.fill();
            }
            
            return;
        }
        
        

        preRenderCanvas.height = radius * 2;
        preRenderCanvas.width = radius * 2;

        gradientFilterCanvas.height = radius * 2;
        gradientFilterCanvas.width = radius * 2;

        var blockingItems = [];
        var i,a,b,c,j;

        if(!isometricMode) {
            for(i = 0; i < lightBlockers.length; i++) {
                a = x - lightBlockers[i].x;
                b = y - lightBlockers[i].y;
                
                c = Math.sqrt( a*a + b*b );

                if(c <= radius + (radius * 0.05)) {
                    lightBlockers[i].d = c;
                    blockingItems.push(lightBlockers[i]);
                }
            }
        }



        if(isometricMode && radius > (relativeGridSize * 0.8)) {
            preRenderContext.save();

            preRenderContext.transform(1,0,0,0.5,0,0);

            radialGradient = ctx.createRadialGradient(radius, radius * 2, 0, radius, radius * 2, radius);
            
            radialGradient.addColorStop(0.0, from);
            radialGradient.addColorStop(1,off);
    
            preRenderContext.fillStyle = radialGradient;
            preRenderContext.beginPath();
            
            preRenderContext.arc(radius, radius * 2, radius, 0, twoPi);

            preRenderContext.fill();
            preRenderContext.restore();

        } else {

            radialGradient = ctx.createRadialGradient(radius, radius, 0, radius, radius, radius);
            
            
            
                    
            radialGradient.addColorStop(0.0, from);
            radialGradient.addColorStop(1,off);
    
            preRenderContext.fillStyle = radialGradient;
            preRenderContext.beginPath();

            preRenderContext.arc(radius, radius, radius, 0, twoPi);
            preRenderContext.fill();
        }
        
        if(isometricMode) {
            ctx.save();
            ctx.globalCompositeOperation = "lighten";
            ctx.drawImage(preRenderCanvas,x - radius, y - radius);

            ctx.restore();
            return;
        }

        var allRays = [];
        var allSides = [];


        for(i = 0; i < 360; i++) {
            var ox0 = radius + (radius + radius * 0.05) * Math.cos(i);
            var oy0 = radius + (radius + radius * 0.05) * Math.sin(i);

            allRays.push({
                x: ox0, y:oy0
            });
        }
        

        
        for(i = 0, j = blockingItems.length; i < j; i++) {
            var item = blockingItems[i];

            var angleZ = Math.atan2(item.y - y, item.x - x);

            var baseX = radius + item.d * Math.cos(angleZ);
            var baseY = radius + item.d * Math.sin(angleZ);


            var topLeft = {x: baseX, y: baseY };
            var topRight = {x: baseX + relativeGridSize, y: baseY };
            var bottomLeft = {x: baseX, y: baseY + relativeGridSize };
            var bottomRight = {x: baseX + relativeGridSize, y: baseY + relativeGridSize };

            allSides.push({
                x1: topLeft.x,
                y1: topLeft.y,
                x2: topRight.x,
                y2: topRight.y,
                source: topLeft
            });

            allSides.push({
                x1: topRight.x,
                y1: topRight.y,
                x2: bottomRight.x,
                y2: bottomRight.y,
                source: topLeft
            });

            allSides.push({
                x1: bottomLeft.x,
                y1: bottomLeft.y,
                x2: bottomRight.x,
                y2: bottomRight.y,
                source: topLeft
            });

            allSides.push({
                x1: topLeft.x,
                y1: topLeft.y,
                x2: bottomLeft.x,
                y2: bottomLeft.y,
                source: topLeft
            });


        }

        var sourceBlocks = [];
        
        for(i = 0, j = allRays.length; i < j; i++) {

            var blocked = false;

            var closestHit = 99999999;
            var closestSource = null;

            
            for(var k = 0, l = allSides.length; k < l; k++) {

                var side = allSides[k];

                var topIntersect = checkLineIntersection(radius,radius,allRays[i].x,allRays[i].y,side.x1,side.y1,side.x2,side.y2);
                


                if(topIntersect.onLine1 && topIntersect.onLine2) {
                    
                    a = radius - topIntersect.x;
                    b = radius - topIntersect.y;
                    
                    c = Math.sqrt( a*a + b*b );

                    if(c < closestHit) {
                        allRays[i].x = topIntersect.x;
                        allRays[i].y = topIntersect.y;
                        closestHit = c;

                        if(side.source) {
                            closestSource = side.source;
                        }
                    }

                }
            }

            if(closestSource != null) {
                sourceBlocks.push(closestSource);
            }

            allRays[i].angle = Math.atan2(allRays[i].y - radius, allRays[i].x - radius);


            if(blocked) {
                continue;
            }

        }

        gradientFilterContext.fillStyle = "rgba(0,0,0,.3)";
        gradientFilterContext.fillRect(0,0,gradientFilterCanvas.width,gradientFilterCanvas.height);

        gradientFilterContext.fillStyle = "#ff0000";

        for(i = 0, j = sourceBlocks.length; i < j; i++) {
            gradientFilterContext.fillRect(sourceBlocks[i].x,sourceBlocks[i].y,relativeGridSize,relativeGridSize);
        }

        allRays.sort(sortRayAngle);

        
        gradientFilterContext.beginPath();
        gradientFilterContext.moveTo(allRays[0].x,allRays[0].y);
        for(i = 1, j = allRays.length; i < j; i++) {
            gradientFilterContext.lineTo(allRays[i].x,allRays[i].y);
        }

        gradientFilterContext.closePath();
        gradientFilterContext.fill();

        

        preRenderContext.save();
        
        preRenderContext.globalCompositeOperation = "destination-in";
        preRenderContext.drawImage(gradientFilterCanvas,0,0);
        preRenderContext.restore();


        ctx.save();
        ctx.globalCompositeOperation = "lighten";
        ctx.drawImage(preRenderCanvas,x - radius, y - radius);

        ctx.restore();

    }

    var relativeGridSize;

    function getZoomLevel() {
        return zoomLevel;
    }

    function setZoomLevel(level) {

        var position = positionToCoords((winW / 2),(winH / 2));


        zoomLevel = level;

        setRelativeGridSize();

        sceneDidChange = true;
        forceSceneChange();

        centerOnCoord(position.x,position.y);
    }

    var halfgs;
    var halfrgs;

    function setRelativeGridSize() {
        relativeGridSize = (0.5 + (zoomLevel * gridSize)) | 0;

        halfrgs = (0.5 + (relativeGridSize / 2)) | 0;
        halfgs = (0.5 + (gridSize / 2)) | 0;
    }
    
    function setFitGridToHeight(fit) {
        fitGridToHeight = fit;
        sizeElements();
    }
    
    function renderFullMap(callback) {
        fullMapRender = true;
        fullRenderCallback = callback;
    }

    function sizeElements() {
        
        if(!canvas) {
            return;
        }

        sceneDidChange = true;
        
        if(usingHolder) {
            winW = canvasHolder.offsetWidth;
            winH = canvasHolder.offsetHeight;
        } else {
            winW = window.innerWidth;
            winH = window.innerHeight;
        }

        if(hdMode) {
            canvas.height = winH * appScale;
            canvas.width = winW * appScale;

            lightCanvas.height = winH * appScale;
            lightCanvas.width = winW * appScale;

            lightFinalRendererCanvas.height = winH * appScale;
            lightFinalRendererCanvas.width = winW * appScale;

            if(usingHolder) {
                staticCanvas.height = winH * appScale;
                staticCanvas.width = winW * appScale;
            }
        } else {
            canvas.height = winH;
            canvas.width = winW;

            lightCanvas.height = winH;
            lightCanvas.width = winW;

            lightFinalRendererCanvas.height = winH;
            lightFinalRendererCanvas.width = winW;

            if(usingHolder) {
                staticCanvas.height = winH;
                staticCanvas.width = winW;
            }
        }
        
        if(fitGridToHeight) {
            gridSize = canvas.height / mapHeight;
        }

        setRelativeGridSize();
        
        mapTotalWidth = mapWidth * relativeGridSize;
        mapTotalHeight = mapHeight * relativeGridSize;

        setLimits();
        setViewBounds();
        sceneDidChange = true;
    }
    
    function setUpdateFunction(func,rate) {
        updateFunction = func;
        
        updateTime = 1000 / rate;
    }
    
    function setIsometric(value) {
        isometricMode = value;
    }
    
    function setSnowing(value) {
        if(particleEngine) {
            particleEngine.snowRunning = value;
        }
    }

    function setEmbers(value) {
        if(particleEngine) {
            particleEngine.embersRunning = value;
        }
    }
    
    function setFixedScale() {
        // will do something here somday maybe idk
    }
    
    function setHDMode(mode) {
        hdMode = mode;
    }
    
    function setLimits() {

        if(isometricMode) {
            xMin = -(mapTotalWidth / 2) - relativeGridSize;
            yMin = 0 - relativeGridSize;
            xLimit = (mapTotalWidth / 2) - winW + (relativeGridSize * 2);
            yLimit = (mapTotalHeight / 2) - winH + (relativeGridSize * 2);
        } else {
            xMin = 0 - (winW / 2);
            yMin = 0 - (winH / 2);
            xLimit = mapTotalWidth - (winW / 2);
            yLimit = mapTotalHeight - (winH / 2);
            
            if(fitGridToHeight) {
                yMin = 0;
                yLimit = 0;
            }
        }

    }

    function setPaintFunction(func) {
        paintFunction = func;
    }
    
    function setMapDims(w,h) {
        mapWidth = w;
        mapHeight = h;

        mapTotalWidth = mapWidth * relativeGridSize;
        mapTotalHeight = mapHeight * relativeGridSize;

        setLimits();
        
        drawQueue = [];

        if(usingHolder) {

            if(staticCallMade) {
                initStaticMap();
            }
        }

    }

    var staticMapInit = false;
    var staticCallMade = false;

    function initStaticMap() {
        staticTilemap = [];

        var mapLength = staticMaxZIndex * mapWidth * mapHeight;

        for(var i = 0; i < mapLength; i++) {
            staticTilemap[i] = null;
        }

        staticMapInit = true;
    }

    function setStaticItem(z,x,y,item) {
        var index = ((x * mapWidth) + y) + (z * (mapWidth * mapHeight));
        staticTilemap[index] = item;
    }

    function getStaticItem(z,x,y) {
        var index = ((x * mapWidth) + y) + (z * (mapWidth * mapHeight));
        return staticTilemap[index];
    }

    function setDrawFunction(func) {
        drawFunction = func;
    }
    
    function setClickFunction(func) {
        clickFunction = func;
    }

    function setRightClickFunction(func) {
        rightClickFunction = func;
    }
    
    function setHoverFunction(func) {
        hoverFunction = func;
    }
    
    function setHoldFunction(func) {
        holdFunction = func;
    }
    
    function setDoubleHeightTiles(dht) {
        doubleHeightTiles = dht;
    }
    
    /**
     * Centers the map on a specific location.  Can be a grid coordinate or an exact pixel position.
     * @param {number} x - x coordinate or exact pixel x location
     * @param {number} y - y coordinate or exact pixel y location
     * @param {boolean} [exactLocation=false] - use exact pixel location
     */
    function centerOnCoord(x,y,exactLocation) {

        sceneDidChange = true;

        if(exactLocation) {
            viewX = x - (winW / 2); 
            viewY = y - (winH / 2);
            
        } else {

            lastGamepadHoverX = x;
            lastGamepadHoverY = y;

            if(isometricMode) {
                viewX = ((x - y) * halfrgs - (winW / 2));
                viewY = ((x + y) * (relativeGridSize / 4) - (winH / 2));
            } else {
                viewX = Math.floor(((x * relativeGridSize) + halfrgs) - (winW / 2)); 
                viewY = Math.floor(((y * relativeGridSize) + halfrgs)- (winH / 2));
            }
        }

        if(fitGridToHeight) {
            viewY = 0;
        }

        checkBounds();

        setViewBounds();

    }
    
    function setViewPadding(padding) {
        viewPadding = padding;
    }

    function checkVisibility(x,y) {
        var dX = x * relativeGridSize;
        var dY = y * relativeGridSize;

        
        
        if(isometricMode) {
            dX = (x - y) * halfrgs;
            dY = (x + y) * (relativeGridSize / 4);
        }

        if(dX > (viewX - (relativeGridSize * viewPadding)) && dX < (winW + (relativeGridSize * viewPadding)) + viewX && dY < (winH + (relativeGridSize * viewPadding)) + viewY && dY > (viewY - (relativeGridSize * viewPadding))) {
            return true;
        } else {
            return false;
        }
        
    }
    
    function setGridSize(size) {

        sceneDidChange = true;

        var centerX = Math.floor(((winW / 2) + viewX) / gridSize);
        var centerY = Math.floor(((winH / 2) + viewY) / gridSize);

        
        
        gridSize = size;

        setRelativeGridSize();
        
        sizeElements();
        
        centerOnCoord(centerX,centerY);
    }
    
    function checkBounds() {
        if(viewX < xMin) {
            viewX = xMin;
        }
        
        if(viewY < yMin) {
            viewY = yMin;
        }
        
        if(viewX > xLimit) {
            viewX = xLimit;
        }
        
        if(viewY > yLimit) {
            viewY = yLimit;
        }

        setViewBounds();
    }

    var lastUpdate;
    var timeDiff;
    var currentFPS;
    
    function getFPS() {

        var len = runningFPS.length;

        var totfps = 0;

        for(var i = 0; i < len; i++) {
            totfps += runningFPS[i];
        }

        var avg = totfps / len;

        return (0.5 + avg) | 0;
    }
    
    var lastUpdateUpdate;
    var updateDiff;

    var usingParticleEnging = false;
    
    function processUpdate(timestamp) {

        lightSources = [];

        if(mouseScrollingUp || mouseScrollingRight || mouseScrollingLeft || mouseScrollingDown) {
            sceneDidChange = true;

            if(mouseScrollingUp) {
                viewY -= 40;
            }

            if(mouseScrollingDown) {
                viewY += 40;
            }

            if(mouseScrollingRight) {
                viewX += 40;
            }

            if(mouseScrollingLeft) {
                viewX -= 40;
            }

            
            checkBounds();
        }

        gamepadNeogtiationTimer++;

        if(gamepadNeogtiationTimer >= 4) {
            gamepadNeogtiationTimer = 0;
            performGamepadUpdate();
        }


        if(!updateFunction || updateFunction == null) {
            return;
        }

        if(isNaN(lastUpdateUpdate)) {
            updateDiff = 1000;
        } else {
            updateDiff = timestamp - lastUpdateUpdate;
        }
        
        if(updateDiff >= updateTime) {
            lastUpdateUpdate = timestamp;
            updateFunction(timestamp);

            if(usingParticleEnging) {
                particleEngine.update(timestamp);
            }
        }

    }

    function forceSceneChange() {
        setMapDims(mapWidth,mapHeight);
        sceneDidChange = true;
    }
    
    var drawQueue;
    var instructionsRecieved = false;
    var lightBlockers = [];
    var renderOrder = 0;

    var runningFPS = [];

    function render(timestamp) {

        window.requestAnimationFrame(render);

        var len;

        renderOrder = 0;

        if(lastUpdate) {
            timeDiff = timestamp - lastUpdate;
            currentFPS = 1000 / timeDiff;
            runningFPS.push(currentFPS);

            if(runningFPS.length > 150) {
                runningFPS.shift();
            }
        }

        lightBlockers = [];
        
        lastUpdate = timestamp;
        
        processUpdate(timestamp);

        for(i = 0, len = textPoppers.length; i < len; i++) {
            if(!textPoppers[i].cashed) {
                textPoppers[i].y -= 2;
                textPoppers[i].alpha -= 0.03;
            }
        }

        if(fullMapRender && fullRenderCallback != null) {
            fullCanvas = document.createElement("canvas");
            fullContext = fullCanvas.getContext("2d");
            fullCanvas.height = mapTotalHeight / 3;
            fullCanvas.width = mapTotalWidth / 3;
        }

        fpscounter++;
        if(fpscounter < fpslimiter) {
            return;
        }

        fpscounter = 0;

        instructionsRecieved = false;

        if(drawFunction) {
            drawFunction();
        }

        if(!instructionsRecieved) {
            return;
        }

        if(usingHolder && sceneDidChange) {

            staticContext.clearRect(0,0,staticCanvas.width,staticCanvas.height);
            staticContext.save();
            staticContext.translate(-viewX,-viewY);

            doStaticRender(staticContext);

            staticContext.restore();

            sceneDidChange = false;
        }

        /*
        if(screenshotPhase) {
            screenshotContext.fillStyle = "#000000";
            screenshotContext.fillRect(0,0,canvas.width,canvas.height);
            screenshotContext.save();
            screenshotContext.translate(-viewX,-viewY);

            doStaticRender(screenshotContext);

            screenshotContext.restore();
        }
        */

        context.clearRect(0,0,canvas.width,canvas.height);

        context.globalCompositeOperation = "source-over";

        context.drawImage(staticCanvas,0,0);

        var z;

        var i;


        if(drawQueue) {


            drawQueue.sort(sortRenderItem);

            while(drawQueue.length) {
                z = drawQueue.shift();
                performRenderOnItem(z,context);

                /*
                if(screenshotPhase) {
                    performRenderOnItem(z,screenshotContext);
                }
                */

                usedRenderObjects.push(z);
            }
        }


        

        

        if(lightSources.length > 0 || filterColor != null) {

            if(filterColor != null && filterAmount > 0) {
                var darkRgb = hexToRgb(filterColor);
                darknessMaskColor = "rgba(" + darkRgb.r + "," + darkRgb.g + "," + darkRgb.b + "," + filterAmount + ")";
            } else {
                darknessMaskColor = null;
            }

            var useOff = "rgba(255,255,255,0)";
            
            if(darknessMaskColor != null) {
                useOff = darknessMaskColor;
            }

            lightContext.fillStyle = useOff;
            lightContext.clearRect(0,0,lightCanvas.width,lightCanvas.height);
            lightContext.fillRect(0,0,lightCanvas.width,lightCanvas.height);

            lightFinalRendererContext.clearRect(0,0,lightFinalRendererCanvas.width,lightFinalRendererCanvas.height);


            while(lightSources.length > 0) {
                var todraw = lightSources.shift();

                var useX = (0.5 + (todraw.x - viewX)) | 0;
                var useY = (0.5 + (todraw.y - viewY)) | 0;

                var radius = todraw.radius;

                ligthenGradient(lightContext, useX, useY, radius, "rgba(" + todraw.color.r + "," + todraw.color.g + "," + todraw.color.b + "," + todraw.intensity + ")", "rgba(" + todraw.color.r + "," + todraw.color.g + "," + todraw.color.b + ",0)");

                lightInstructionRecycling.push(todraw);
            }

            lightFinalRendererContext.globalCompositeOperation = "source-over";
            lightFinalRendererContext.drawImage(canvas,0,0);
            lightFinalRendererContext.globalCompositeOperation = "source-in";
            lightFinalRendererContext.drawImage(lightCanvas,0,0);

            context.save();
            context.globalCompositeOperation = "multiply";
            context.drawImage(lightFinalRendererCanvas,0,0);
            context.restore();

        }

        if(usingParticleEnging) {
            particleEngine.draw(context);

            /*
            if(screenshotPhase) {
                particleEngine.draw(screenshotContext);
            }
            */
        }




        if(screenshotPhase) {
            
            screenshotPhase = false;
            sceneDidChange = true;

            screenshotContext.drawImage(canvas,0,0);

            if(screenshotCallback != null) {
                screenshotCallback(screenshotCanvas);
                
            }

            screenshotCanvas = null;
            screenshotContext = null;
            screenshotCallback = null;
            
        }


        var popperKillArray = [];
        
        // render text poppers
        for(i = 0, len = textPoppers.length; i < len; i++) {
            if(textPoppers[i].cashed) {
                popperKillArray.push(textPoppers[i]);
            } else {
                renderTextPopper(context,textPoppers[i]);
            }
        }

        for(i = 0, len = popperKillArray.length; i < len; i++) {
            usedPopperObjects.push(popperKillArray[i]);
            ParticleEngine2.removeFromArray(textPoppers,popperKillArray[i]);
        }

        popperKillArray = null;



        if(isSelection && !isPinching) {
            context.save();
            context.globalAlpha = 1;
            context.lineWidth = 4;
            context.strokeStyle = selectionColor;
            context.strokeRect(selectionTopX,selectionTopY,selectionWidth,selectionHeight);
            context.restore();
        }
        
        if(fullMapRender && fullRenderCallback != null) {

            var fullMap = new Image();
            fullMap.src = fullCanvas.toDataURL("image/jpeg",0.4);
            fullRenderCallback(fullMap);
            
            fullMapRender = false;
            fullRenderCallback = null;
            fullCanvas = null;
            fullContext = null;


        }
    }


    var screenshotCanvas = null;
    var screenshotContext = null;
    var screenshotPhase = false;
    var screenshotCallback = null;

    function doStaticRender(ctx) {

        if(!staticCallMade) {
            return;
        }

        if(currentViewBounds == null) {
            setViewBounds();
        }

        var minX = currentViewBounds.minX;
        var maxX = currentViewBounds.maxX;
        var minY = currentViewBounds.minY;
        var maxY =  currentViewBounds.maxY;

        if(fullMapRender) {
            minX = 0;
            maxX = mapWidth;
            minY = 0;
            maxY = mapHeight;
        }

        for(var z = 0; z <= staticMaxZIndex; z++) {
            for(var x = minX; x <= maxX; x++) {
                for(var y = minY; y <= maxY; y++) {
                    
                    var renderItem = getStaticItem(z,x,y);

                    if(renderItem == null) {
                        continue;
                    }

                    ctx.drawImage(renderItem.img,renderItem.x - 1,renderItem.y,renderItem.w + 2,renderItem.h + 1);

                    if(fullMapRender && fullContext !== undefined) {
                        fullContext.drawImage(renderItem.img,Math.floor(renderItem.x * gridSize) / 3,((renderItem.y - 1) * gridSize) / 3,(gridSize / 3),((gridSize * 2) / 3));
                    }
                    
                }
            }
        }
    }

    function takeScreenshot(callback) {

        if(!canvas) {
            callback();
            return;
        }

        screenshotCanvas = document.createElement("canvas");
        screenshotContext = screenshotCanvas.getContext("2d");

        screenshotCanvas.width = canvas.width;
        screenshotCanvas.height = canvas.height;

        screenshotCallback = callback;
        screenshotPhase = true;
    }

    function performRenderOnItem(z,ctx) {

        var useX = (0.5 + (z.x - viewX)) | 0;
        var useY = (0.5 + (z.y - viewY)) | 0;

        

        if(z.blocksLight) {

            var lbY = useY;

            if(doubleHeightTiles) {
                lbY += relativeGridSize;
            }

            if(!isometricMode) {
                lightBlockers.push({
                    x: useX,
                    y: lbY
                });
            }
        }

        if(z.isLine) {
            var useX2 = (0.5 + (z.w - viewX)) | 0;
            var useY2 = (0.5 + (z.h - viewY)) | 0;

            ctx.strokeStyle = z.color;
            ctx.globalAlpha = 1;
            screenshotCanvas = 
            ctx.beginPath();
            

            ctx.moveTo(useX,useY);
            ctx.lineTo(useX2,useY2);

            ctx.stroke();
        } else {
            if(z.isSquare) {
                ctx.fillStyle = z.color;
                ctx.globalAlpha = 1;
                
                if(isometricMode) {
                    ctx.beginPath();
                    ctx.moveTo(useX,useY);

                    ctx.lineTo(useX + halfrgs,(useY + (relativeGridSize / 4)));
                    ctx.lineTo(useX,(useY + halfrgs));
                    ctx.lineTo(useX - halfrgs,(useY + (relativeGridSize / 4)));
                    
                    ctx.fill();
                } else {
                    ctx.fillRect(useX,useY,relativeGridSize,relativeGridSize);
                }
                
            } else {
                if(z.isBar) {
                    var barWidth = z.barWidth;
                    var meterWidth = z.meterWidth;
                    
                    if(meterWidth < 0) {
                        meterWidth = 0;
                    }
                    
                    ctx.globalAlpha = 1;
                    
                    ctx.fillStyle = "#000000";
                    ctx.fillRect(useX,useY,barWidth,6);
                    
                    ctx.fillStyle = z.barColor;
                    ctx.fillRect(useX + 1,useY + 1,meterWidth,4);

                } else {

                    if(z.alpha && z.alpha != null) {
                        ctx.globalAlpha = z.alpha;
                    } else {
                        ctx.globalAlpha = 1;
                    }
                    
                    if(z.segmentX >= 0) {

                        ctx.drawImage(z.img,z.segmentX,0,halfgs,z.sH,useX + z.destSegX,useY,halfrgs,z.h);
                        


                    } else {
                        if(zoomLevel == 1) {
                            ctx.drawImage(z.img,useX,useY);
                        } else {
                            if(z && z.img) {
                                ctx.drawImage(z.img,useX,useY,z.w,z.h);
                            }
                        }
                        
                        if(fullMapRender && fullContext && fullContext != null) {
                            fullContext.drawImage(z.img,Math.floor(useX * gridSize) / 3,((useY - 1) * gridSize) / 3,(gridSize / 3),((gridSize * 2) / 3));
                        }
                    }
                    
                }
            }
        }
        
        
    }

    function getFreshRenderItem() {

        var ri;

        if(usedRenderObjects.length == 0) {
            ri = new RenderItem();
        } else {
            ri = usedRenderObjects.shift();
            ri.x = 0;
            ri.y = 0;
            ri.img = null;
            ri.w = 0;
            ri.h = 0;
            ri.alpha = null;
            ri.isBar = false;
            ri.barWidth = 0;
            ri.meterWidth = 0;
            ri.barColor = "#000000";
            ri.isSquare = false;
            ri.isLine = false;
            ri.color = "#000000";
            ri.nearness = 0;
            ri.tileWidth = 1;
            ri.tileHeight = 1;
            ri.zIndex = 0;
            ri.altitude = 0;
            ri.segmentX = -1;
            ri.cX = 0;
            ri.cY = 0;
            ri.destSegX = 0;
            ri.sW = 0;
            ri.sH = 0;
            ri.blocksLight = true;
        }

        ri.ro = renderOrder;

        renderOrder++;

        return ri;
    }

    function RenderItem() {
        this.x = 0;
        this.y = 0;
        this.img = null;
        this.w = 0;
        this.h = 0;
        this.alpha = null;
        this.isBar = false;
        this.barWidth = 0;
        this.meterWidth = 0;
        this.barColor = "#000000";
        this.isSquare = false;
        this.isLine = false;
        this.color = "#000000";
        this.nearness = 0;
        this.tileWidth = 1;
        this.tileHeight = 1;
        this.zIndex = 0;
        this.altitude = 0;
        this.segmentX = -1;
        this.cX = 0;
        this.cY = 0;
        this.destSegX = 0;
        this.sW = 0;
        this.sH = 0;
        this.blocksLight = true;
        this.ro = 0;
    }

    function Chevron(color,image) {
        if(color) {
            this.color = color;
        } else {
            this.color = null;
        }

        if(image) {
            this.image = image;
        } else {
            this.image = null;
        }
    }

    function sortRayAngle(a,b) {
        if(a.angle > b.angle) {
            return 1;
        }

        if(a.angle < b.angle) {
            return -1;
        }
    }

    function sortRenderItem(a,b) {
        if(a.zIndex > b.zIndex) {
            return 1;
        }

        if(a.zIndex < b.zIndex) {
            return -1;
        }

        if(a.nearness > b.nearness) {
            return 1;
        }

        if(a.nearness < b.nearness) {
            return -1;
        }

        if(a.cX > b.cX) {
            return 1;
        }

        if(a.cX < b.cX) {
            return -1;
        }
        

        if(a.cY > b.cY) {
            return 1;
        }

        if(a.cY < b.cY) {
            return -1;
        }

        if(a.ro > b.ro) {
            return 1;
        }

        if(a.ro < b.ro) {
            return -1;
        }

        return 0;
    }

    function drawBigStationary(img,x,y,tX,tY,step,meterPercent,meterColor,alpha,zIndex,tileWidth,tileHeight,chevrons) {

        var dX;
        var dY;
        var dW = img.width;
        var dH = img.height;

        if(zoomLevel != 1) {

            dW *= zoomLevel;
            dH *= zoomLevel;

        }

        var lastTileX = x + tileWidth - 1;
        var lastTileY = y + tileHeight - 1;
        
        var bottomIsoY = (lastTileX + lastTileY) * (relativeGridSize / 4);
        var rightIsoX = (lastTileX - y) * halfrgs;

        var totalBottom = bottomIsoY + halfrgs;
        var totalRight = rightIsoX + halfrgs;

        dX = totalRight - dW;
        dY = totalBottom - dH;


        // this is faster in theory, but it might have been what was causeing last line issue on iOS 
        // var drawSegments = (0.5 + (img.width / gridSize)) | 0;
        var drawSegments = Math.round(img.width / halfgs);

        var segX = 0;
        var dSegX = 0;

        var segmentSize = halfgs * zoomLevel;

        var z;

        for(var i = 0; i < drawSegments; i++) {
            z = getFreshRenderItem();

            var segLeftX = dX + (segmentSize * i);
            var segRightX = segLeftX + segmentSize;
            var nearness = x + y;

            // to get nearness for segment, look at all tiles in the object and get the biggest Y for tiles that appear within the segment 
            for(var sX = x; sX < x + tileWidth; sX++) {
                
                for(var sY = y; sY < y + tileHeight; sY++) {

                    var isoX = (sX - sY) * halfrgs;

                    var tDX = isoX - halfrgs;

                    var right = isoX + halfrgs;

                    if(right > totalRight) {
                        right = totalRight;
                    }

                    var tileNearness = sX + sY;

                    if(right >= segLeftX && tDX <= segRightX) {
                        if(tileNearness > nearness) {
                            nearness = tileNearness;
                        }
                    }
                }

            }

            if(!meterColor || meterColor == null) {
                meterColor = "#ff0000";
            }

            z.sW = img.width;
            z.sH = img.height;


            z.img = img;

            z.zIndex = zIndex;

            z.cX = x;
            z.cY = y;

            z.x = dX;
            z.y = dY;
            z.w = dW;
            z.h = dH;

            if(tileWidth) {
                z.tileWidth = tileWidth;
            }

            if(tileHeight) {
                z.tileHeight = tileHeight;
            }

            z.nearness = nearness;

            if(alpha && alpha != null) {
                z.alpha = alpha;
            }

            z.segmentX = segX;
            z.destSegX = dSegX;

            segX += halfgs;
            dSegX += halfrgs;

            drawQueue.push(z);

            if(i == tileWidth - 1) {
                var hadBar = false;


                if(meterPercent && meterPercent != null) {
                    hadBar = true;

                    z = getFreshRenderItem();
                    z.isBar = true;
                    z.x = dX ;
                    z.y = dY + dH;
                    z.barWidth = dW;
                    z.meterWidth = (dW * meterPercent) - 2;
                    z.barColor = meterColor;

                    z.zIndex = zIndex + 1;

                    z.nearness = nearness + 1;
                    if(isometricMode) {
                        z.nearness = nearness + 2;
                    }

                    drawQueue.push(z);

                }

                if(chevrons != null) {
                    
                    var cPos = dX;
                    var yPos = dY + dH;
        
                    if(hadBar) {
                        yPos += 10;
                    }
        
                    for(var cv = 0; cv < chevrons.length; cv++) {
                        z = getFreshRenderItem();
                        z.isBar = true;
                        z.x = cPos;
                        z.y = yPos;
                        z.barWidth = 6;
                        z.meterWidth = 4;
                        z.barColor = chevrons[cv].color;
            
                        z.nearness = nearness + 1;
            
                        if(isometricMode) {
                            z.nearness = nearness + 2;
                        }
            
                        z.zIndex = zIndex + 1;
            
                        drawQueue.push(z);
        
                        cPos += 10;
                    }
                }
            }

            
        }

        var loc = {
            "x":    dX,
            "y":    dY,
            "w":    dW,
            "h":    dH
        };
        
        return loc;
    }
    
    function drawSprite(img,x,y,tX,tY,step,meterPercent,meterColor,alpha,zIndex,tileWidth,tileHeight,chevrons) {
        
        if(!zIndex) {
            zIndex = 0;
        }

        if(!chevrons) {
            chevrons = null;
        }
        
        if(!checkVisibility(x,y) && !fullMapRender) {
            return;
        }

        

        instructionsRecieved = true;

        if(tX == null) {
            tX = x;
            tY = y;
        }

        if(isometricMode && tX == x && tY == y && !useFastRenderMode) {
            if(tileWidth && tileHeight && (tileWidth > 1 || tileHeight > 1)) {
                return drawBigStationary(img,x,y,tX,tY,step,meterPercent,meterColor,alpha,zIndex,tileWidth,tileHeight,chevrons);
            } else {
                return drawBigStationary(img,x,y,tX,tY,step,meterPercent,meterColor,alpha,zIndex,1,1,chevrons);
            }
        }  

        
        var z = getFreshRenderItem();
        
        z.blocksLight = false;

        if(!meterColor || meterColor == null) {
            meterColor = "#ff0000";
        }

        var dX;
        var dY;
        var dW = img.width;
        var dH = img.height;

        if(zoomLevel != 1) {

            dW *= zoomLevel;
            dH *= zoomLevel;

            dW = (0.5 + dW) | 0;
            dH = (0.5 + dH) | 0;
        }

        var nearness = y;
        
        if(isometricMode) {

            nearness += x;

            var isoX = (x - y) * halfrgs;
            var isoY = (x + y) * (relativeGridSize / 4);

            if(tileWidth && tileHeight) {

                var lastTileX = x + tileWidth - 1;
                var lastTileY = y + tileHeight - 1;
                
                var bottomIsoY = (lastTileX + lastTileY) * (relativeGridSize / 4);
                var rightIsoX = (lastTileX - y) * halfrgs;

                var totalBottom = bottomIsoY + halfrgs;
                var totalRight = rightIsoX + halfrgs;

                dX = totalRight - dW;
                dY = totalBottom - dH;
            } else {
                if(tileWidth) {
                    dX = isoX - ((relativeGridSize * tileWidth) / 2);
                } else {

                    var imgGridWidth = dW / relativeGridSize;

                    dX = isoX - ((relativeGridSize * imgGridWidth) / 2);
                }

                var bottomLeftY = isoY + halfrgs;
                
                dY = bottomLeftY - dH;
            }

            


        } else {

            dX = x * relativeGridSize;
            dY = ((y * relativeGridSize) + relativeGridSize) - dH;

        }
        
        if(x != tX || y != tY) {
            var tenthGrid = relativeGridSize / 20;
            var stepDistance = tenthGrid * step;


            if(isometricMode) {

                var nX = x;
                var nY = y;

                if(tX > x) {
                    dX += stepDistance / 2;
                    dY += stepDistance / 4;

                    nX += 2;
                }
                
                if(tX < x) {
                    dX -= stepDistance / 2;
                    dY -= stepDistance / 4;

                    nX++;
                }
                
                if(tY > y) {
                    dX -= stepDistance / 2;
                    dY += stepDistance / 4;

                    nY += 2;
                }
                
                if(tY < y) {
                    dX += stepDistance / 2;
                    dY -= stepDistance / 4;

                    nY++;
                }

                nearness = nX + nY;

            } else {
                if(tX > x) {
                    dX += stepDistance;
                }
                
                if(tX < x) {
                    dX -= stepDistance;
                }
                
                if(tY > y) {
                    dY += stepDistance;
                }
                
                if(tY < y) {
                    dY -= stepDistance;
                }
            }
            
        }
        
        z.img = img;

        z.zIndex = zIndex;

        z.x = dX;
        z.y = dY;
        z.w = dW;
        z.h = dH;

        if(tileWidth) {
            z.tileWidth = tileWidth;
        }

        if(tileHeight) {
            z.tileHeight = tileHeight;
        }

        z.nearness = nearness;

        if(alpha && alpha != null) {
            z.alpha = alpha;
        }

        drawQueue.push(z);

        var hadBar = false;

        if(meterPercent && meterPercent != null) {
            hadBar = true;

            z = getFreshRenderItem();
            z.isBar = true;
            z.x = dX;
            z.y = dY + dH;
            z.barWidth = dW;
            z.meterWidth = (dW * meterPercent) - 2;
            z.barColor = meterColor;

            z.nearness = nearness + 1;

            if(isometricMode) {
                z.nearness = nearness + 2;
            }

            z.zIndex = zIndex + 1;

            drawQueue.push(z);

        }

        if(chevrons != null) {

            var cPos = dX;
            var yPos = dY + dH;

            if(hadBar) {
                yPos += 10;
            }

            for(var cv = 0; cv < chevrons.length; cv++) {
                z = getFreshRenderItem();
                z.isBar = true;
                z.x = cPos;
                z.y = yPos;
                z.barWidth = 6;
                z.meterWidth = 4;
                z.barColor = chevrons[cv].color;
    
                z.nearness = nearness + 1;
    
                if(isometricMode) {
                    z.nearness = nearness + 2;
                }
    
                z.zIndex = zIndex + 1;
    
                drawQueue.push(z);

                cPos += 10;
            }
        }

        var loc = {
            "x":    dX,
            "y":    dY,
            "w":    dW,
            "h":    dH
        };
        
        return loc;
    }
    
    function drawStaticTile(img,x,y,zIndex,blocksLight) {

        if(img == undefined) {
            return;
        }

        if(x < 0 || x > mapWidth) {
            return;
        }

        if(y < 0 || y > mapHeight) {
            return;
        }

        if(!zIndex) {
            zIndex = 0;
        }

        if(zIndex > staticMaxZIndex) {
            staticMaxZIndex = zIndex;
            initStaticMap();
            sceneDidChange = true;
        }

        if(!usingHolder) {
            drawTile(img,x,y,zIndex);
            return;
        }

        if(!checkVisibility(x,y) && !fullMapRender) {
            return;
        }

        staticCallMade = true;

        if(!staticMapInit) {
            initStaticMap();
        }

        instructionsRecieved = true;

        var existingItem = getStaticItem(zIndex,x,y);

        if(existingItem == null) {
            
            setStaticItem(zIndex,x,y, buildRenderObjectForTile(img,x,y));
            sceneDidChange = true;
        } else {
            if(existingItem.img != img) {
                
                setStaticItem(zIndex,x,y, buildRenderObjectForTile(img,x,y,existingItem));
                sceneDidChange = true;
            }
        }

    }

    function drawTile(img,x,y,zIndex,blocksLight) {

        if(img == undefined) {
            return;
        }

        if(blocksLight == undefined) {
            blocksLight = false;
            
        }

        instructionsRecieved = true;

        if(!zIndex) {
            zIndex = 0;
        }
        
        if(!checkVisibility(x,y) && !fullMapRender) {
            return;
        }

        var z = buildRenderObjectForTile(img,x,y);

        z.nearness = y;

        z.blocksLight = blocksLight;

        if(isometricMode) {
            z.nearness += x;
        }

        z.zIndex = zIndex;

        drawQueue.push(z);
    }

    function buildRenderObjectForTile(img,x,y,z) {
        if(z == undefined || z == null) {
            z = getFreshRenderItem();
        }

        z.img = img;

        var stdWidth = img.width;
        var stdHeight = img.height;

        if(zoomLevel != 1) {
            stdWidth *= zoomLevel;
            stdHeight *= zoomLevel;

        }

        if(isometricMode) {
            var isoX = (x - y) * halfrgs;
            var isoY = (x + y) * (relativeGridSize / 4);
            
            var bottomLeftX = isoX - halfrgs;
            var bottomLeftY = isoY + halfrgs;
            
            var drawY = bottomLeftY - stdHeight;

            var width = stdWidth;
            var height = stdHeight;

            z.x = bottomLeftX;
            z.y = drawY;
            z.w = width;
            z.h = height;
            z.nearness = x + y;
        } else {
            

            var stdX = x * relativeGridSize;
            var stdY = ((y * relativeGridSize) + relativeGridSize) - stdHeight;

            z.x = stdX;
            z.y = stdY;
            z.w = stdWidth;
            z.h = stdHeight;

        }

        return z;
    }

    function drawLine(color,x,y,x2,y2,zIndex) {
        if(!zIndex) {
            zIndex = 0;
        }

        var nearness = (x * 2) + (y * 2) + zIndex;

        var z = getFreshRenderItem();
        z.isLine = true;
        z.color = color;

        if(isometricMode) {
            
            var isoX = ((x - y) * halfrgs) + halfrgs;
            var isoY = ((x + y) * (relativeGridSize / 4)) + (relativeGridSize / 4);
            
            z.x = isoX;
            z.y = isoY;

            isoX = ((x2 - y2) * halfrgs) + halfrgs;
            isoY = ((x2 + y2) * (relativeGridSize / 4)) + (relativeGridSize / 4);
            
            z.w = isoX;
            z.h = isoY;

        } else {
            
            z.x = (x * relativeGridSize) + halfrgs;
            z.y = (y * relativeGridSize) + halfrgs;

            z.w = (x2 * relativeGridSize) + halfrgs;
            z.h = (y2 * relativeGridSize) + halfrgs;

        }

        z.nearness = nearness;

        drawQueue.push(z);
    }
    
    function drawSquare(color,x,y,zIndex) {

        if(!zIndex) {
            zIndex = 0;
        }
        
        if(!checkVisibility(x,y) && !fullMapRender) {
            return;
        }

        var nearness = y;

        var z = getFreshRenderItem();
        z.isSquare = true;
        z.color = color;

        if(isometricMode) {

            nearness += x;
            
            var isoX = (x - y) * halfrgs;
            var isoY = (x + y) * (relativeGridSize / 4);
            
            z.x = isoX;
            z.y = isoY;

        } else {
            
            z.x = x * relativeGridSize;
            z.y = y * relativeGridSize;

        }

        z.nearness = nearness;
        z.zIndex = zIndex;

        drawQueue.push(z);

    }
    
    function popText(x,y,text,color) {

        if(!color || color == null) {
            color = "#ff0000";
        }
        
        var popX;
        var popY;
        
        if(isometricMode) {
            var isoX = (x - y) * halfrgs;
            var isoY = (x + y) * (relativeGridSize / 4);
            
            popX = (isoX - halfrgs) + halfrgs;
            var bottomLeftY = isoY + halfrgs;
            
            popY = (bottomLeftY - relativeGridSize) + halfrgs;
        } else {
            popX = x * relativeGridSize + halfrgs;
            popY = y * relativeGridSize + halfrgs;
        }

        var freshPop;
        
        if(usedPopperObjects.length == 0) {
            freshPop = new TextPopItem(popX,popY,text,color);
        } else {
            freshPop = usedPopperObjects.shift();
            freshPop.x = popX;
            freshPop.y = popY;
            freshPop.text = text;
            freshPop.color = color;
            freshPop.cashed = false;
            freshPop.alpha = 1.0;
        }

        textPoppers.push(freshPop);
    }
    
    function TextPopItem(x,y,text,color) {
        this.x = x;
        this.y = y;
        this.text = text;
        this.color = color;
        this.cashed = false;
        this.alpha = 1.0;

    }

    function renderTextPopper(ctx,popper) {

        if(popper.alpha <= 0) {
            popper.cashed = true;
            return;
        }

        var useX = popper.x - viewX;
        var useY = popper.y - viewY;

        var fontSize = 20 * zoomLevel;

        ctx.strokeStyle = "#000000";

        ctx.fillStyle = popper.color;
        ctx.textAlign = "center";
        ctx.textBaseline = "middle";
        ctx.font = "bold " + fontSize + "px sans-serif";
        
        ctx.save();
        ctx.globalAlpha = popper.alpha;
        
        ctx.fillText(popper.text,useX,useY);
        ctx.strokeText(popper.text,useX,useY);
        
        ctx.restore();

    }

    function getOffset(el) {
        el = el.getBoundingClientRect();
        return {
            left: el.left + window.scrollX,
            top: el.top + window.scrollY
        };
    }
    
    var tHX;
    var tHY;

    var pointersDown = 0;
    var allPointers = {};
    var primaryPointer = null;
    var lastDistance = null;

    function getPointer(id,x,y,type) {
        if(allPointers[id]) {
            return allPointers[id];
        }

        allPointers[id] = {};

        allPointers[id].x = x;
        allPointers[id].y = y;
        allPointers[id].lastX = x;
        allPointers[id].lastY = y;
        allPointers[id].origX = x;
        allPointers[id].origY = y;
        allPointers[id].primary = false;
        allPointers[id].type = type;
        allPointers[id].rightclick = false;

        if(pointersDown == 0) {
            allPointers[id].primary = true;
            primaryPointer = allPointers[id];
        }

        return allPointers[id];
    }


    function getUsingTouch() {
        return usingTouch;
    }

    function down(object,id,x,y,type,pressure,which) {

        if(type == "touch") {
            usingTouch = true;
        } else {
            usingTouch = false;
        }

        var pointer = getPointer(id,x,y,type);
        pointer.down = true;


        pointersDown++;

        isSelection = false;
        
        actualX = x;
        actualY = y;

        if(type == "mouse") {
            isUsingMouse = true;

            if(which && which == 3) {
                pointer.rightclick = true;
            } else {
                pointer.rightclick = false;
            }

        } else {
            isUsingMouse = false;
        }
        
        if(usingHolder) {
            x -= getOffset(canvas).left;
            y -= getOffset(canvas).top;
        }

        startX = x;
        startY = y;
        didMove = false;
        isDown = true;
        isHold = false;

        
        
        if(holdFunction || selectionFunction) {

            tHX = x;
            tHY = y;

            if(!pointer.rightclick) {
                holdTimer = setTimeout(handleTheHold,700);
            }
        }

        if(pointersDown > 1) {
            isSelection = false;
            clearTimeout(holdTimer);
        }
    }

    function handleTheHold() {
        
        didMove = false;

        if(isPainting) {
            return;
        }
        

        if(selectionFunction) {
            isSelection = true;
            selectionX = tHX;
            selectionY = tHY;

            setSelectionPoints(tHX,tHY);

            if(selectionProgressFunction) {
                sendSelectionPointsToFunction(selectionProgressFunction);
            }

            return;
        }

        isDown = false;
        isHold = true;

        var position = positionToCoords(hx,hy);

        holdFunction(position.x,position.y);
    }

    function Coord(x,y) {
        this.x = x;
        this.y = y;
    }

    function positionToCoords(x,y) {

        var fX,fY;

        if(isometricMode) {
            fX = Math.floor(((x + viewX) / halfrgs + (y + viewY) / (relativeGridSize / 4)) /2);
            fY = Math.floor(((y + viewY) / (relativeGridSize / 4) - (x + viewX) / halfrgs) /2);
        } else {
            fX = Math.floor((x + viewX) / relativeGridSize);
            fY = Math.floor((y + viewY) / relativeGridSize);
        }

        return new Coord(fX,fY);
    }
    
    var paintTimer;

    var paintTimeoutFunction;

    function playParticleEffect(effect,x,y) {
        if(!usingParticleEnging) {
            return;
        }

        var px,py;

        if(isometricMode) {
            px = (x - y) * halfrgs;
            py = (x + y) * (relativeGridSize / 4);
        } else {
            px = x * relativeGridSize; 
            py = y * relativeGridSize;
        }

        var useX = px - viewX;
        var useY = py - viewY;

        particleEngine.playEffect(effect,useX,useY);
    }
    
    function alertPainting(callback,timeout) {

        if(!timeout) {
            timeout = 600;
        }

        if(callback) {
            paintTimeoutFunction = callback;
        } else {
            paintTimeoutFunction = null;
        }

        isPainting = true;
        
        if(paintTimer) {
            clearTimeout(paintTimer);
            paintTimer = null;
        }
        
        paintTimer = setTimeout(clearOutPaint,timeout);
    }

    function clearOutPaint(noCallback) {
        isPainting = false;
        paintTimer = null;

        if(noCallback) {
            return;
        }

        if(paintTimeoutFunction != null) {
            paintTimeoutFunction();
        }
    }

    var isUsingMouse = false;

    var actualX;
    var actualY;

    var lastGamepadHoverX = 0;
    var lastGamepadHoverY = 0;

    
    function move(object,id,x,y,type,pressure,which) {



        if(type == "touch") {
            usingTouch = true;
        } else {
            usingTouch = false;
        }

        var pointer = getPointer(id,x,y,type);
        pointer.down = true;


        actualX = x;
        actualY = y;

        if(type == "mouse") {
            isUsingMouse = true;

            
        } else {
            isUsingMouse = false;
        }
        
        if(usingHolder) {
            x -= getOffset(canvas).left;
            y -= getOffset(canvas).top;
        }

        pointer.x = x;
        pointer.y = y;

        if(pointersDown == 2) {
            didMove = true;
            handlePinch();
            return;
        }

        var oldX = 0;
        var oldY = 0;
        
        if(curX) {
            oldX = curX;
            oldY = curY;
        }
        
        curX = x;
        curY = y;

        var position = positionToCoords(x,y);

        if((type == "mouse" || type =="pen") && hoverFunction) {

            lastGamepadHoverX = position.x;
            lastGamepadHoverY = position.y;

            hoverFunction(position.x,position.y,actualX,actualY,type);
        }
        
        if(!isDown) {
            return;
        }

        if(isSelection) {

            setSelectionPoints(x,y);

            if(selectionProgressFunction) {
                sendSelectionPointsToFunction(selectionProgressFunction);
            }

            return;
        }
        
        if(didMove) {

            sceneDidChange = true;
            
            if(isPainting && paintFunction) {

                if(lastPaintX != null) {
                    if(hx == lastPaintX && hy == lastPaintY) {
                        return;
                    }
                }

                lastPaintX = position.x;
                lastPaintY = position.y;

                paintFunction(position.x,position.y,actualX,actualY);
                return;
            }
            
            difX = oldX - curX;
            difY = oldY - curY;
            
            viewX += difX;
            viewY += difY;
            
            checkBounds();
        } else {
            if(x > startX + 12 || x < startX - 12 || y > startY + 12 || y < startY - 12) {
                didMove = true;
                if(holdTimer) {
                    isHold = false;
                    clearTimeout(holdTimer);
                }
            }
        }
    }

    var lastPaintX = null;
    var lastPaintY = null;

    function setSelectionPoints(x,y) {
        if(x < selectionX) {
            selectionTopX = x;
            selectionWidth = selectionX - x;
        } else {
            selectionTopX = selectionX;
            selectionWidth = x - selectionX;
        }

        if(y < selectionY) {
            selectionTopY = y;
            selectionHeight = selectionY - y;
        } else {
            selectionTopY = selectionY;
            selectionHeight = y - selectionY;
        }
    }

    function modifyViewX(amount) {

        sceneDidChange = true;

        viewX += amount;
        checkBounds();
    }

    function modifyViewY(amount) {

        sceneDidChange = true;
        
        viewY += amount;
        checkBounds();
    }
    
    function up(object,id,type,which) {

        if(type == "touch") {
            usingTouch = true;
        } else {
            usingTouch = false;
        }

        var pointer = getPointer(id,null,null,type);

        if(pointer.down) {
            pointersDown--;
            pointer.down = false;
        }





        if(holdTimer) {
            isHold = false;
            clearTimeout(holdTimer);
        }

        if(pointer.primary) {
            primaryPointer = null;
            pointer.primary = false;
        }

        if(isPinching) {
            isSelection = false;
        }

        isPinching = false;

        if(!isDown) {
            return;
        }

        if(isHold) {
            return;
        }

        isPainting = false;

        if(pointersDown <= 0) {
            allPointers = {};
            pointersDown = 0;
            primaryPointer = null;
            lastDistance = null;
        }

        if(isSelection) {
            processSelection();
            return;
        }
        

        
        if(!didMove && clickFunction) {

            var position = positionToCoords(startX,startY);

            isSelection = false;
            isDown = false;

            tcCx = position.x;
            tcCy = position.y;
            tcAx = actualX;
            txAy = actualY;

            if(type == "touch") {
                setTimeout(function() {
                    doClick("touch");
                },75);
            } else {
                if(pointer.rightclick) {
                    doRightClick(type);
                    pointer.rightclick = false;
                } else {
                    doClick(type);
                }
                
            }
        }
        
        isDown = false;
        didMove = false;

        actualX = null;
        actualY = null;

        
    }

    var tcCx;
    var tcCy;
    var tcAx;
    var txAy;

    function doRightClick(type) {
        
        pointersDown = 0;

        if(!rightClickFunction) {
            return;
        }

        rightClickFunction(tcCx,tcCy,tcAx,txAy,type);
    }

    function doClick(type) {

        pointersDown = 0;

        if(!clickFunction) {
            return;
        }

        clickFunction(tcCx,tcCy,tcAx,txAy,type);
    }

    // https://github.com/facebook/fixed-data-table/blob/master/src/vendor_upstream/dom/normalizeWheel.js
    // Reasonable defaults
    var PIXEL_STEP  = 10;
    var LINE_HEIGHT = 40;
    var PAGE_HEIGHT = 800;

    function normalizeWheel(/*object*/ event) /*object*/ {
        var sX = 0, sY = 0,       // spinX, spinY
            pX = 0, pY = 0;       // pixelX, pixelY

        // Legacy
        if ("detail"      in event) { sY = event.detail; }
        if ("wheelDelta"  in event) { sY = -event.wheelDelta / 120; }
        if ("wheelDeltaY" in event) { sY = -event.wheelDeltaY / 120; }
        if ("wheelDeltaX" in event) { sX = -event.wheelDeltaX / 120; }

        // side scrolling on FF with DOMMouseScroll
        if ( "axis" in event && event.axis === event.HORIZONTAL_AXIS ) {
            sX = sY;
            sY = 0;
        }

        pX = sX * PIXEL_STEP;
        pY = sY * PIXEL_STEP;

        if ("deltaY" in event) { pY = event.deltaY; }
        if ("deltaX" in event) { pX = event.deltaX; }

        if ((pX || pY) && event.deltaMode) {
            if (event.deltaMode == 1) {          // delta in LINE units
                pX *= LINE_HEIGHT;
                pY *= LINE_HEIGHT;
            } else {                             // delta in PAGE units
                pX *= PAGE_HEIGHT;
                pY *= PAGE_HEIGHT;
            }
        }

        // Fall-back if spin cannot be determined
        if (pX && !sX) { sX = (pX < 1) ? -1 : 1; }
        if (pY && !sY) { sY = (pY < 1) ? -1 : 1; }

        return { 
            pinX  : sX,
            spinY  : sY,
            pixelX : pX,
            pixelY : pY 
        };
    }


    function onMouseWheel(e) {

        e.preventDefault();

        if(!isUsingMouse) {
            return;
        }

        var normalized = normalizeWheel(e);

        if(e.deltaY < 0) {
            doZoom(true,normalized.spinY);
        }

        if(e.deltaY > 0) {
            doZoom(false,normalized.spinY);
        }

        
    }

    function doZoom(zoomIn,amount) {

        var lvl = zoomLevel;

        if(amount) {

            if(amount > 0.15) {
                amount = 0.15;
            }

            if(amount < -0.15) {
                amount = -0.15;
            }

            if(zoomIn) {
                if(lvl < 3) {
                    lvl -= (lvl * amount);
                }
            } else {
                if(lvl > 0.1) {
                    lvl -= (lvl * amount);
                }
            }
        } else {
            if(zoomIn) {
                if(lvl < 3) {
                    lvl += (lvl * 0.15);
                }
            } else {
                if(lvl > 0.1) {
                    lvl -= (lvl * 0.15);
                }
            }
        }

        if(lvl > 3) {
            lvl = 3;
        }

        if(lvl < 0.1) {
            lvl = 0.1;
        }

        setZoomLevel(lvl);
    }

    var isPinching = false;

    function handlePinch() {
        if(pointersDown != 2 || primaryPointer == null) {
            return;
        }

        var pointer1,pointer2;

        for(var prop in allPointers) {
            if(allPointers[prop].down) {
                if(!pointer1) {
                    pointer1 = allPointers[prop];
                } else {
                    if(!pointer2) {
                        pointer2 = allPointers[prop];
                    }
                }
            }
        }

        if(!pointer1 || !pointer2) {
            return;
        }

        isSelection = false;
        clearTimeout(holdTimer);
        isPinching = true;

        var a = pointer1.x - pointer2.x;
        var b = pointer1.y - pointer2.y;

        var c = Math.sqrt( a*a + b*b );

        if(lastDistance == null) {
            lastDistance = c;
            return;
        }


        if(c > lastDistance + 3) {
            doZoom(true);
            lastDistance = c;
        }

        if(c < lastDistance - 3) {
            doZoom(false);
            lastDistance = c;
        }

    }

    function setSelectionFunction(fct,progress,color) {

        if(color) {
            selectionColor = color;
        }

        if(progress) {
            selectionProgressFunction = progress;
        }

        selectionFunction = fct;
    }

    function processSelection() {
        isSelection = false;
        isDown = false;
        sendSelectionPointsToFunction(selectionFunction);
    }

    function sendSelectionPointsToFunction(func) {

        var rx = selectionTopX + selectionWidth;
        var by = selectionTopY + selectionHeight;

        var topLeft = getCoordAndActual(selectionTopX,selectionTopY);
        var topRight = getCoordAndActual(rx,selectionTopY);
        var bottomLeft = getCoordAndActual(selectionTopX,by);
        var bottomRight = getCoordAndActual(rx,by);

        func(topLeft,topRight,bottomLeft,bottomRight);
    }

    function getCoordAndActual(actualX,actualY) {
        var result = {};

        result.aX = actualX;
        result.aY = actualY;

        var position = positionToCoords(actualX,actualY);

        result.x = position.x;
        result.y = position.y;

        return result;
    }

    var mouseScrollingLeft = false;
    var mouseScrollingRight = false;
    var mouseScrollingUp = false;
    var mouseScrollingDown = false;

    function startPan(direction) {
        if(direction == "up") {
            mouseScrollingUp = true;
        }

        if(direction == "down") {
            mouseScrollingDown = true;
        }

        if(direction == "left") {
            mouseScrollingLeft = true;
        }

        if(direction == "right") {
            mouseScrollingRight = true;
        }
    }

    function endPan(direction) {
        if(direction == "up") {
            mouseScrollingUp = false;
        }

        if(direction == "down") {
            mouseScrollingDown = false;
        }

        if(direction == "left") {
            mouseScrollingLeft = false;
        }

        if(direction == "right") {
            mouseScrollingRight = false;
        }
    }


    function desktopWindowPointerMove(event) {

        if(!parent.document.fullscreenElement && !parent.document.webkitFullscreenElement && !parent.document.mozFullScreenElement && !parent.document.msFullscreenElement) {
            return;
        }

        if (event.pointerType && event.pointerType != "mouse") {
            return;
        }

        if(event.pageX < 6) {
            mouseScrollingLeft = true;
        } else {
            mouseScrollingLeft = false;
        }

        if(event.pageX > window.innerWidth - 6) {
            mouseScrollingRight = true;
        } else {
            mouseScrollingRight = false;
        }

        if(event.pageY < 6) {
            mouseScrollingUp = true;
        } else {
            mouseScrollingUp = false;
        }

        if(event.pageY > window.innerHeight - 6) {
            mouseScrollingDown = true;
        } else {
            mouseScrollingDown = false;
        }
    }

    var gamepadNeogtiationTimer = 0;

    var gpLeftDown = false;
    var gpUpDown = false;
    var gpRightDown = false;
    var gpDownDown = false;

    var gpRLeftDown = false;
    var gpRUpDown = false;
    var gpRRightDown = false;
    var gpRDownDown = false;


    var aDownTime = 0;

    var ltDown = false;
    var rtDown = false;

    //var altTurn = false;

    function gamepadDown(button) {

        if(button == "lt") {
            ltDown = true;
        }

        if(button == "rt") {
            rtDown = true;
        }

        if(button == "a") {

            aDownTime = new Date().getTime();
        }

        if(button == "up") {
            gpUpDown = true;
        }

        if(button == "down") {
            gpDownDown = true;
        }

        if(button == "left") {
            gpLeftDown = true;
        }

        if(button == "right") {
            gpRightDown = true;
        }


        if(button == "rup") {
            gpRUpDown = true;
        }

        if(button == "rdown") {
            gpRDownDown = true;
        }

        if(button == "rleft") {
            gpRLeftDown = true;
        }

        if(button == "rright") {
            gpRRightDown = true;
        }
    }

    function gamepadUp(button) {

        if(button == "lt") {
            ltDown = false;
        }

        if(button == "rt") {
            rtDown = false;
        }

        if(button == "a") {

            var aTime = aDownTime - new Date().getTime();

            if(aTime < 400 && clickFunction) {
                clickFunction(lastGamepadHoverX,lastGamepadHoverY,0,0,"gamepad");
            }
        }

        if(button == "up") {
            gpUpDown = false;
        }

        if(button == "down") {
            gpDownDown = false;
        }

        if(button == "left") {
            gpLeftDown = false;
        }

        if(button == "right") {
            gpRightDown = false;
        }


        if(button == "rup") {
            gpRUpDown = false;
        }

        if(button == "rdown") {
            gpRDownDown = false;
        }

        if(button == "rleft") {
            gpRLeftDown = false;
        }

        if(button == "rright") {
            gpRRightDown = false;
        }
    }

    function performGamepadUpdate() {

        var didHover = false;

        // up
        if(gpUpDown && !gpRightDown && !gpLeftDown && !gpDownDown) {
            lastGamepadHoverY--;
            
            if(isometricMode) {
                lastGamepadHoverX--;
            }

            didHover = true;
        }


        // up right
        if(gpUpDown && gpRightDown && !gpLeftDown && !gpDownDown) {
            lastGamepadHoverY--;
            
            if(!isometricMode) {
                lastGamepadHoverX++;
            }

            didHover = true;
        }

        // right
        if(!gpUpDown && gpRightDown && !gpLeftDown && !gpDownDown) {
            lastGamepadHoverX++;
            
            if(isometricMode) {
                lastGamepadHoverY--;
            }

            didHover = true;
        }

        // down right
        if(!gpUpDown && gpRightDown && !gpLeftDown && gpDownDown) {
            lastGamepadHoverX++;
            
            if(!isometricMode) {
                lastGamepadHoverY++;
            }

            didHover = true;
        }

        // down
        if(!gpUpDown && !gpRightDown && !gpLeftDown && gpDownDown) {
            lastGamepadHoverY++;
            
            if(isometricMode) {
                lastGamepadHoverX++;
            }

            didHover = true;
        }

        // down left
        if(!gpUpDown && !gpRightDown && gpLeftDown && gpDownDown) {
            lastGamepadHoverY++;
            
            if(!isometricMode) {
                lastGamepadHoverX--;
            }

            didHover = true;
        }


        // left
        if(!gpUpDown && !gpRightDown && gpLeftDown && !gpDownDown) {
            lastGamepadHoverX--;
            
            if(isometricMode) {
                lastGamepadHoverY++;
            }

            didHover = true;
        }

        // up left
        if(gpUpDown && !gpRightDown && gpLeftDown && !gpDownDown) {
            lastGamepadHoverX--;
            
            if(!isometricMode) {
                lastGamepadHoverY++;
            }

            didHover = true;
        }

        if(didHover && hoverFunction) {
            hoverFunction(lastGamepadHoverX,lastGamepadHoverY,0,0,"gamepad");
            centerOnCoord(lastGamepadHoverX,lastGamepadHoverY);
        }

        var didScroll = false;

        if(gpRLeftDown) {
            viewX -= 40;
            sceneDidChange = true;
            didScroll = true;
        }

        if(gpRRightDown) {
            viewX += 40;
            sceneDidChange = true;
            didScroll = true;
        }

        if(gpRUpDown) {
            viewY -= 40;
            sceneDidChange = true;
            didScroll = true;
        }

        if(gpRDownDown) {
            viewY += 40;
            sceneDidChange = true;
            didScroll = true;
        }

        if(didScroll) {
            checkBounds();
        }

        if(rtDown && !ltDown) {
            doZoom(true);
        }

        if(!rtDown && ltDown) {
            doZoom(false);
        }
    }

    function handleInputNonWac(object,downFunction,moveFunction,endFunction) {
        
        object.addEventListener("contextmenu",function() {
            return false;
        });
        
        //Use either pointer events or touch/mouse combo
        if (window.PointerEvent) {
            

            object.touchAction = "none";
            object.contentZooming = "none";
            object.msTouchAction = "none";
            object.msContentZooming = "none";
            
            object.style.touchAction = "none";
            
            
            
            object.addEventListener("pointerdown", function (event) {
                event.preventDefault();
                
                var pointerType = "mouse";
                
                if (event.pointerType) {
                    pointerType = event.pointerType;
                }
                
                var pressure = 1;
                
                if (event.pressure) {
                    pressure = event.pressure;
                }

                downFunction(object, event.pointerId, event.pageX, event.pageY, pointerType, pressure);
                return false;
            },false);

            object.addEventListener("pointermove", function (event) {
                event.preventDefault();

                var pointerType = "mouse";
                
                if (event.pointerType) {
                    pointerType = event.pointerType;
                }
                
                var pressure = 1;
                
                if (event.pressure) {
                    pressure = event.pressure;
                }

                moveFunction(object, event.pointerId, event.pageX, event.pageY, pointerType, pressure);
                return false;
            },false);

            object.addEventListener("pointerout", function (event) {

                var pointerType = "mouse";
                
                if (event.pointerType) {
                    pointerType = event.pointerType;
                }

                endFunction(object, event.pointerId, pointerType);
                return false;
            },false);

            object.addEventListener("pointerup", function (event) {

                var pointerType = "mouse";
                
                if (event.pointerType) {
                    pointerType = event.pointerType;
                }

                endFunction(object, event.pointerId, pointerType);
                return false;
            },false);
        } else {
            object.addEventListener("mousedown",function(event) {
                event.preventDefault();
                downFunction(object,1,event.pageX,event.pageY,"mouse",1);
                return false;
            },false);
            
            object.addEventListener("mousemove",function(event) {
                event.preventDefault();
                moveFunction(object,1,event.pageX,event.pageY,"mouse",1);
                return false;
            },false);
            
            object.addEventListener("mouseout",function(event) {
                event.preventDefault();
                endFunction(object,1,"mouse");
                return false;
            },false);
            
            object.addEventListener("mouseup",function(event) {
                event.preventDefault();
                endFunction(object,1,"mouse");
                return false;
            },false);
            
            if ("ontouchstart" in window) {
                object.addEventListener("touchstart",function(event) {
                    event.preventDefault();
                    
                    for(var i=0;i<event.changedTouches.length;i++) {
                        downFunction(object,event.changedTouches[i].identifier,event.changedTouches[i].pageX,event.changedTouches[i].pageY,"touch",1);
                    }
                    return false;
                },false);
                
                object.addEventListener("touchmove",function(event) {
                    event.preventDefault();
                    
                    for(var i=0;i<event.changedTouches.length;i++) {
                        moveFunction(object,event.changedTouches[i].identifier,event.changedTouches[i].pageX,event.changedTouches[i].pageY,"touch",1);
                    }
                    
                    return false;
                },false);
                
                object.addEventListener("touchend",function(event) {
                    event.preventDefault();
                    
                    for(var i=0;i<event.changedTouches.length;i++) {
                        endFunction(object,event.changedTouches[i].identifier,"touch");
                    }
                    
                    return false;
                },false);
                
                object.addEventListener("touchleave",function(event) {
                    event.preventDefault();
                    
                    for(var i=0;i<event.changedTouches.length;i++) {
                        endFunction(object,event.changedTouches[i].identifier,"touch");
                    }
                    return false;
                },false);
                
                object.addEventListener("touchcancel",function(event) {
                    event.preventDefault();
                    
                    for(var i=0;i<event.changedTouches.length;i++) {
                        endFunction(object,event.changedTouches[i].identifier,"touch");
                    }
                    return false;
                },false);
            }
        }
    }

    function ViewBounds(minX,minY,maxX,maxY) {
        this.minX = minX;
        this.minY = minY;
        this.maxX = maxX;
        this.maxY = maxY;

        if(this.minX < 0) {
            this.minX = 0;
        }

        if(this.minY < 0) {
            this.minY = 0;
        }

        if(this.maxX > mapWidth) {
            this.maxX = mapWidth;
        }

        if(this.maxY > mapHeight) {
            this.maxY = mapHeight;
        }
    }

    function setViewBounds() {

        var min = positionToCoords(0 - viewPadding,0 - viewPadding);
        var max = positionToCoords(winW + viewPadding,winH + viewPadding);

        if(isometricMode) {

            var tr = positionToCoords(winW + viewPadding,0 - viewPadding);
            var bl = positionToCoords(0 - viewPadding,winH + viewPadding);

            currentViewBounds = new ViewBounds(min.x,tr.y,max.x,bl.y);
        } else {
            currentViewBounds = new ViewBounds(min.x,min.y,max.x,max.y);
        }

    }

    function getViewBounds() {
        if(currentViewBounds == null) {
            setViewBounds();
        }

        return currentViewBounds;
    }

    function checkLineIntersection(line1StartX, line1StartY, line1EndX, line1EndY, line2StartX, line2StartY, line2EndX, line2EndY) {
        var denominator, a, b, numerator1, numerator2, result = {
            x: null,
            y: null,
            onLine1: false,
            onLine2: false
        };
        denominator = ((line2EndY - line2StartY) * (line1EndX - line1StartX)) - ((line2EndX - line2StartX) * (line1EndY - line1StartY));
        if (denominator == 0) {
            return result;
        }
        a = line1StartY - line2StartY;
        b = line1StartX - line2StartX;
        numerator1 = ((line2EndX - line2StartX) * a) - ((line2EndY - line2StartY) * b);
        numerator2 = ((line1EndX - line1StartX) * a) - ((line1EndY - line1StartY) * b);
        a = numerator1 / denominator;
        b = numerator2 / denominator;
    
        result.x = line1StartX + (a * (line1EndX - line1StartX));
        result.y = line1StartY + (a * (line1EndY - line1StartY));

        if (a > 0 && a < 1) {
            result.onLine1 = true;
        }

        if (b > 0 && b < 1) {
            result.onLine2 = true;
        }

        return result;
    }    
    
})();