/* global webAppCore */
/* global ColonyGame */
(function() {

    window.wacMenuOptionSelected = onMenuOptionSelected;

    window.addEventListener("load",startApp);

    var currentModFile = null;
    var currentFileName = null;
    var pendingOpenFileName = null;

    function startApp() {
        webAppCore.startApp(initPage);
    }

    function initPage() {

        currentModFile = new TextureModData();

        linkLayouts();
        setupMenuOptions();

        populateTextureList(true);
        populateMetaData();

        var lastSaveData = webAppCore.getVar("text-edit-last-save","0");

        if(lastSaveData == "0") {
            webAppCore.clearSplash();
        } else {
            pendingOpenFileName = lastSaveData;
            webAppCore.openFile(lastSaveData,onFileOpened);
        }

    }

    function TextureModData() {
        this.meta = {
            "name": "",
            "revision": 0,
            "author": "",
            "description": "",
            "url": ""
        };

        this.overrides = {};
    }

    function linkLayouts() {
        document.getElementById("btnOptions").addEventListener("click",function() {
            webAppCore.toggleMenu();
        });

        var metaButtons = document.querySelectorAll("[data-meta-item]");

        for(var i = 0; i < metaButtons.length; i++) {
            metaButtons[i].addEventListener("change",function() {
                var prop = this.dataset.metaItem;
                currentModFile.meta[prop] = this.value.latinize();
            });
        }

        document.getElementById("overrideImportImage").onclick = function() {
            webAppCore.importImage(onImageImport);
        };

        document.getElementById("clearOverride").onclick = function() {
            if(currentAssetItem == null) {
                return;
            }

            delete currentModFile.overrides[currentAssetItem.name];

            loadOverrideForCurrent();
            populateTextureList(false);
        };


        
    }

    function setupMenuOptions() {
        webAppCore.clearMenu();

        webAppCore.addMenuOption("Close Editor","wacMenuOptionSelected","back","#F44336","closeeditor");

        webAppCore.addMenuOption("Save","wacMenuOptionSelected","save",null,"save");

        webAppCore.addMenuOption("Open","wacMenuOptionSelected","openfile",null,"open");

        webAppCore.addMenuOption("Export","wacMenuOptionSelected","savelocal",null,"export");
    }

    function onMenuOptionSelected(name,tag) {

        if(tag == "closeeditor") {
            if(currentFileName == null) {
                var moditems = 0;

                for(var prop in currentModFile.overrides) {
                    console.log(prop);
                    moditems++;
                }

                if(moditems > 0) {
                    webAppCore.showDialog({
                        "title":    "Unsaved Changes",
                        "message":  "There are unsaved changes on this mod.  Do you want to save your work before leaving?",
                        "op1": {
                            "text":     "Save",
                            "func":     function() {
                                saveMod(closeEditor);
                            }
                        },
                        "op2": {
                            "text":     "Quit",
                            "func":     closeEditor
                        }
                    });
                } else {
                    closeEditor();
                }
            } else {
                saveMod(closeEditor);
            }
        }

        if(tag == "save") {
            saveMod(placeholder);
        }

        if(tag == "open") {
            webAppCore.openListing(onListingBack);
        }

        if(tag == "export") {
            saveMod(function() {
                webAppCore.exportFile(currentFileName,JSON.stringify(currentModFile),false);
            });
        }
    }

    function onListingBack(data) {

        var menuOptions = [];

        for(var i = 0; i < data.length; i++) {
            if(endsWith(data[i],"." + ColonyGame.meta.graphicsModFileFormat)) {
                menuOptions.push({
                    "title": data[i],
                    "tag": data[i]
                });
            }
        }

        if(menuOptions.length == 0) {
            webAppCore.showToast("No installed mods!");
        } else {
            webAppCore.presentNativeList("Open",menuOptions,function(tag) {

                if(!tag) {
                    return;
                }

                pendingOpenFileName = tag;
                webAppCore.openFile(tag,onFileOpened);
            },webAppCore.getString("general-dismiss"));
        }
    }

    function endsWith(str, suffix) {
        
        if(!str || !suffix) {
            return false;
        }

        return str.indexOf(suffix, str.length - suffix.length) !== -1;
    }

    function closeEditor() {
        webAppCore.showSplash("Saving mod...");
        webAppCore.goBack();
    }

    function populateTextureList(loadFirst) {
        document.getElementById("textureListing").innerHTML = "";

        for(var i = 0; i < ColonyGame.tiles.length; i++) {
            var sisebarItem = document.createElement("p");
            sisebarItem.innerHTML = ColonyGame.tiles[i].name;
            sisebarItem.textureReference = ColonyGame.tiles[i];
            sisebarItem.className = "sidebarItem";

            if(currentModFile.overrides[ColonyGame.tiles[i].name]) {
                sisebarItem.className += " override";
            }

            sisebarItem.addEventListener("click",function() {
                loadTexture(this.textureReference);
            });

            document.getElementById("textureListing").appendChild(sisebarItem);

            if(loadFirst && i == 0) {
                loadTexture(ColonyGame.tiles[i]);
            }
        }
    }

    function populateMetaData() {
        var metaButtons = document.querySelectorAll("[data-meta-item]");

        for(var i = 0; i < metaButtons.length; i++) {
            metaButtons[i].value = currentModFile.meta[metaButtons[i].dataset.metaItem];
        }
    }

    var currentOriginal = null;
    var currentAssetItem = null;
    var currentOverride = null;

    function loadTexture(texture) {
        currentAssetItem = texture;

        document.getElementById("curTexName").innerHTML = currentAssetItem.name;
        document.getElementById("curTexWidth").innerHTML = "Frame Width: " + currentAssetItem.tileWidth + "px";

        currentOriginal = new Image();
        currentOriginal.onload = function() {
            var frames = this.width / currentAssetItem.tileWidth;
            document.getElementById("origFrameCount").innerHTML = "Frames: " + frames;
        };
        currentOriginal.src = "images/" + currentAssetItem.image;

        document.getElementById("originalImage").src = "images/" + currentAssetItem.image;

        document.getElementById("origExportImage").onclick = function() {
            var exCanvas = document.createElement("canvas");
            document.body.appendChild(exCanvas);
            exCanvas.width = currentOriginal.width;
            exCanvas.height = currentOriginal.height;
            var exContext = exCanvas.getContext("2d");
            exContext.drawImage(currentOriginal,0,0);

            webAppCore.saveScreenshot(exCanvas);

            setTimeout(function() {
                document.body.removeChild(exCanvas);
            },2000);
            
        };

        loadOverrideForCurrent();
    }

    function clearOverride() {
        document.getElementById("overrideImage").src = "";
        currentOverride = null;
        document.getElementById("overrideFrameCount").innerHTML = "";
    }

    function loadOverrideForCurrent() {
        if(currentAssetItem == null) {
            clearOverride();
            return;
        }

        if(!currentModFile.overrides[currentAssetItem.name]) {
            clearOverride();
            return;
        }

        currentOverride = new Image();
        currentOverride.onload = function() {
            var frames = this.width / currentAssetItem.tileWidth;
            document.getElementById("overrideFrameCount").innerHTML = "Frames: " + frames;
        };
        currentOverride.src = currentModFile.overrides[currentAssetItem.name];

        document.getElementById("overrideImage").src = currentModFile.overrides[currentAssetItem.name];

        
    }

    function onImageImport(data) {
        if(currentAssetItem == null) {
            return;
        }

        var checkImage = new Image();
        checkImage.onload = function() {

            var frames = this.width / currentAssetItem.tileWidth;

            if(isFloat(frames)) {
                webAppCore.showAlert("Image is incorrect size.  Frame width must be equal to the original asset.");
            } else {
                currentModFile.overrides[currentAssetItem.name] = data;
                
                loadOverrideForCurrent();
                populateTextureList(false);
            }

            
        };
        checkImage.src = data;
        

        
    }

    function isFloat(n) {
        return n === +n && n !== (n|0);
    }

    function placeholder() {}
    
    function saveMod(callback) {
        if(currentFileName == null) {
            if(currentModFile.meta.name.trim().length == 0) {
                webAppCore.showAlert("Please choose a name for this mod before saving.");
                return;
            } else {
                currentFileName = cleanFileName(currentModFile.meta.name.trim()) + "." + ColonyGame.meta.graphicsModFileFormat;
            }
        }

        currentModFile.meta.revision++;

        webAppCore.storeVar("text-edit-last-save",currentFileName);
        webAppCore.storeVar("mod-name-ref-" + currentFileName,currentModFile.meta.name.trim());
        webAppCore.storeVar("mod-author-ref-" + currentFileName,currentModFile.meta.author.trim());

        setTimeout(function() {
            var saveContent = JSON.stringify(currentModFile);
            webAppCore.saveFile(currentFileName,saveContent,callback,true);
        },180);

        
    }

    function cleanFileName(name) {
        var clean = name.replace(/([^a-z0-9]+)/gi, "-");
        return clean.toLowerCase();
    }

    function onFileOpened(data) {

        if(data == null || data.trim().length == 0) {
            webAppCore.clearSplash();
            return;
        }

        try {
            currentModFile = JSON.parse(data);
            populateTextureList(true);
            populateMetaData();

            if(pendingOpenFileName != null) {
                currentFileName = pendingOpenFileName;
                webAppCore.storeVar("text-edit-last-save",pendingOpenFileName);
                pendingOpenFileName = null;
            }
        } catch(e) {
            console.log(e);
            currentModFile = null;
        }

        webAppCore.clearSplash();
    }
})();