﻿/* global Android */
/* global html2canvas */
window.addEventListener("resize",sizeElements,false);
window.addEventListener("load",initApp,false);
window.addEventListener("error",onGeneralError,false);

var textToShare = "";
var tmpOutputDiv;
var currentFrameURL;
var listScrollTimeout = null;
var appFrame;
var appIsInit = false;
var currentOpenDialog;

var lastTheme = null;

function initApp() {

    console.log("starting init");

    processAnalytics();
    linkLayouts();
    platformInit();
}

function force720() {
    console.log("forcing 720");
    document.getElementById("viewportMeta").content = "initial-scale=1.5, user-scalable=no";
}

function platformInit() {
    juiceUpIAP();
    completeInit();
}

function completeInit() {
    if(appIsInit) {
        return;
    }

    audioArray = [];
    loopingAudio = document.createElement('audio');

    console.log("completing init");

    if(window.autoclearSplash) {
        setTimeout(clearSplash,2000);
    }

    prepareTitleScreen();
    gooseUpKeyboard();

    appFrame.src = window.appUrl;
    window.appUrl = appFrame.src;

    appIsInit = true;
    sizeElements();

    console.log("finished init");
}

function gooseUpKeyboard() {

    window.addEventListener("keydown",function(event) {

        if(activeDialog != null || document.getElementById("introScreen").style.display == "block") {
            event.preventDefault();

            if(event.keyCode == 38 || event.keyCode == 87 || event.keyCode == 104) {
                onGamepadDown("keyboard","up");
            }

            if(event.keyCode == 40 || event.keyCode == 83 || event.keyCode == 98) {
                onGamepadDown("keyboard","down");
            }

            if(event.keyCode == 37 || event.keyCode == 65 || event.keyCode == 100) {
                onGamepadDown("keyboard","left");
            }

            if(event.keyCode == 39 || event.keyCode == 68 || event.keyCode == 102) {
                onGamepadDown("keyboard","right");
            }

            if(event.keyCode == 13 || event.keyCode == 16 || event.keyCode == 32) {
                onGamepadDown("keyboard","a");
            }

            if(event.keyCode == 27) {
                onGamepadDown("keyboard","start");
            }
        }

    },false);

    window.addEventListener("keyup",function(event) {

        if(activeDialog != null || document.getElementById("introScreen").style.display == "block") {
            event.preventDefault();

            if(event.keyCode == 38 || event.keyCode == 87 || event.keyCode == 104) {
                onGamepadUp("keyboard","up");
            }

            if(event.keyCode == 40 || event.keyCode == 83 || event.keyCode == 98) {
                onGamepadUp("keyboard","down");
            }

            if(event.keyCode == 37 || event.keyCode == 65 || event.keyCode == 100) {
                onGamepadUp("keyboard","left");
            }

            if(event.keyCode == 39 || event.keyCode == 68 || event.keyCode == 102) {
                onGamepadUp("keyboard","right");
            }

            if(event.keyCode == 13 || event.keyCode == 16 || event.keyCode == 32) {
                onGamepadUp("keyboard","a");
            }

            if(event.keyCode == 27) {
                onGamepadUp("keyboard","start");
            }
        }

    },false);
}

function onGamepadDown(id,key) {
    gamepadDown(key,id);
}

function onGamepadUp(id,key) {
    gamepadUp(key,id);
}

function gamepadDown(key,id) {

    if(appFrame.contentWindow.webAppCore == null) {
        return;
    }

    pendingGamepadDirectionKey = key;
    pendingGamepadId = id;

    // check if we are handling the title screen
    if(document.getElementById("introScreen").style.display == "block") {
        titlescreenGamepadDown(key);
        return;
    }

    appFrame.contentWindow.webAppCore.recieveGamepadDown(key,id);
}

function gamepadUp(key,id) {

    if(appFrame.contentWindow.webAppCore == null) {
        return;
    }

    killScrollTimeout();
    appFrame.contentWindow.webAppCore.recieveGamepadUp(key,id);
}

function notifyURL(url) {
    currentFrameURL = url;
    notifyPlatformNavigation();
}

function onPause() {
    if(!appFrame) {
        return;
    }

    if(appFrame.contentWindow.webAppCore != null && appFrame.contentWindow.webAppCore.onPause != null) {
        appFrame.contentWindow.webAppCore.onPause();
    }
}

function onResume() {

    if(!appFrame) {
        appIsInit = false;
        initApp();
        return;
    }

    if(appFrame.contentWindow.webAppCore != null && appFrame.contentWindow.webAppCore.onResume != null) {
        appFrame.contentWindow.webAppCore.onResume();
    }
}

function onBack() {
    appFrame.contentWindow.webAppCore.onBack();
}

function linkLayouts() {
    appFrame = document.getElementById("appFrame");

    if(appFrame != null) {
        appFrame.addEventListener("error",onFrameError,false);
    }
}

function toggleMenu() {
    Android.toggleMenu();
}


function sizeElements() {

    if(!appIsInit) {
        return;
    }

    if(appFrame == undefined) {
        appFrame = document.getElementById("appFrame");
    }

}

var activeDialog;

function showColorPicker(defCol) {

    if(!defCol) {
        defCol = lastTheme;
    }

    Android.pickColor(defCol);
}

function onMenuPress() {
	if(!appFrame || !appFrame.contentWindow.webAppCore) {
		return;
	}
	
    appFrame.contentWindow.webAppCore.revieveMenuPress();
}

function colorSelected(color) {
    appFrame.contentWindow.webAppCore.colorSelected(color);
}

function wacMidiPlaybackComplete() {
	if(!appFrame || !appFrame.contentWindow.webAppCore) {
		return;
	}
	
    appFrame.contentWindow.webAppCore.midiCallbackListener();
}

function addMenuOption(name,icon,color,tag) {
    Android.addNavbarItem(name,icon,color,tag);
}

function navItemSelection(name,tag) {
    appFrame.contentWindow.webAppCore.executeMenuFunction(name,tag);
}

function resetCustomMenu() {
    Android.clearNavbar();
}

function forceCloseMenu() {
    // hrmm
}

function clearSplash() {
    clearPlatformSplash();
}

function saveScreenshot(element) {
    if(!element) {
        element = appFrame.contentWindow.document.body;
    }

    html2canvas(element, {
        onrendered: function(canvas) {
            processScreenshot(canvas);
        }
    });
}

function generateDate() {
    var today = new Date();
    var dd = today.getDate();
    var mm = today.getMonth()+1; //January is 0!

    var yyyy = today.getFullYear();
    if(dd<10){
        dd='0'+dd
    }
    if(mm<10){
        mm='0'+mm
    }
    return mm+'/'+dd+'/'+yyyy;
}

function logEvent(category,action,label) {
    Android.trackEvent(category,action,label, false);
}

function closeOrConfirmAlert() {
    //for legacy purposes
}

function wacChoice1() {
     appFrame.contentWindow.webAppCore.choice1();
}

 function wacChoice2() {
     appFrame.contentWindow.webAppCore.choice2();
 }

function wacInputReturn(text) {
    if(inputCallbackFunction) {
        inputCallbackFunction(text);
    } else {
        appFrame.contentWindow.webAppCore.inputReturn(text);
    }
}

var inputCallbackFunction = undefined;

function inputBox(message,func,isScreenshot) {

    inputCallbackFunction = func;
    Android.inputBox(message);
}

function finalizePremiumAndRemoveBannerAds() {
    storeVar(window.webAppId + "_premium_upgrade","1");
    appFrame.contentWindow.webAppCore.onPremiumUpgraded();
}

// Web App Core v3.1 Functions
function addToolbarItem(name,tag,icon,inSecondarySection) {
    Android.addToolbarItem(name,tag,icon,inSecondarySection);
}

function resetCustomToolbar() {
    Android.resetToolbar();
}

function toggleToolbar() {
    Android.toggleToolbar();
}

function toolbarItemSelected(name,tag) {
    appFrame.contentWindow.webAppCore.executeToolbarFunction(name,tag);
}

function setTitle(title) {
    Android.setTitle(title);
}

// WAC 3.3
var selectedTitleScreenOption = -1;

function prepareTitleScreen() {

    if(Android.getPlatform() == "2") {
        document.getElementById("newsfeed").style.display = "none";
    } else {
        getNewsfeed(document.getElementById("newsfeed"));
    }

    if(Android.getPlatform() == "2" || window.isTvDevice) {
        document.getElementById("apeLogo").style.display = "none";
    }

    document.getElementById("titleLogo").addEventListener("click",function() {
        toggleShareBox("Check out " + window.appName + "! #appurl");
    },false);



    if(window.isTvDevice) {
        document.getElementById("featuredApp").style.display = "none";
    } else {
        fetchFeaturedImage(document.getElementById("featuredApp"));
    }

}

var featureHolder;

function fetchFeaturedImage(output) {
    if(getIsPremium()) {
        output.style.display = "none";
        return;
    }

    if(Android.getIsTvDevice()) {
        output.style.display = "none";
        return;
    }

    featureHolder = output;

    var platformId = Android.getPlatform();

    if(platformId == "2") {
        output.style.display = "none";
        return;
    }

    var apeMarketId = Android.getApeMarketId();
    var buster = 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
        var r = Math.random() * 16 | 0, v = c == 'x' ? r : (r & 0x3 | 0x8);
        return v.toString(16);
    });

    var scoreData = new Array();
    scoreData[scoreData.length] = {key:"p",value:platformId};
    scoreData[scoreData.length] = {key:"a",value:apeMarketId};
    scoreData[scoreData.length] = {key:"zzz",value:buster};

    dataRequest("https://market.ape-apps.com/app_resources/featured_app.php",showFeature,scoreData,false);
}

function showFeature(data) {
    if(!featureHolder || featureHolder == null) {
        return;
    }

    console.log("parsing feature");

    var imageURL = data.split("###");

    featureHolder.src = "https://market.ape-apps.com/app_icons/" + imageURL[0];
    featureHolder.linkRef = imageURL[1];
    featureHolder.onclick = function () {
        var featurePrexif = Android.getFeaturePrefix();

        loadURL(featurePrexif + this.linkRef);
    };

}

function getNewsfeed(output) {
    if(!output) {
        return;
    }

    tmpOutputDiv = output;

    dataRequest("https://market.ape-apps.com/app_resources/title-newsfeed.php",printNewsfeed,null,false);
}

function printNewsfeed(data) {
    if(!tmpOutputDiv || tmpOutputDiv == null) {
        return;
    }

    if (data.indexOf("###") == -1) {
        tmpOutputDiv.innerHTML = data;
    } else {
        var newsParts = [];
        newsParts = data.split("###");
        tmpOutputDiv.innerHTML = newsParts[0];
        tmpOutputDiv.addEventListener("click",function(event) {
            event.preventDefault();
            loadURL(newsParts[1]);
        },false);
        tmpOutputDiv.style.cursor = "pointer";
    }

    try {
        tmpOutputDiv.start();
    } catch(e) {
        //ohwell
    }
    tmpOutputDiv = null;
}

function setTitleVisible(visible) {

    document.getElementById("introFader").style.display = "none";

    if(visible) {
        selectedTitleScreenOption = -1;
        setTitleOptionSelection();
        document.getElementById("introScreen").style.display = "block";
        document.getElementById("newsfeed").start();
    } else {
        document.getElementById("introScreen").style.display = "none";
        appFrame.contentWindow.focus();
    }
}

function setTitleOptionSelection() {
    var titleOptions = document.getElementById("introScreenOptions").getElementsByClassName("introScreenOption");

    for(var i = 0; i < titleOptions.length; i++) {
        if(i == selectedTitleScreenOption) {
            titleOptions[i].style.color = titleOptions[i].selectedColor;
            titleOptions[i].style.fontSize = "32px";
            titleOptions[i].style.fontWeight = "bold";
        } else {
            titleOptions[i].style.color = titleOptions[i].standardColor;
            titleOptions[i].style.fontSize = "26px";
            titleOptions[i].style.fontWeight = "normal";
        }
    }
}

function setTitleBackground(bgHTML) {
    var hasbg = bgHTML.indexOf("url('");
    if(hasbg > -1) {
        var n = bgHTML.indexOf("://");
        if(n == -1) {
            bgHTML = bgHTML.replace("url('","url('app/");
        }
    }

    document.getElementById("introScreen").style.background = bgHTML;
}

function setTitleLogo(logo) {
    var n = logo.indexOf("://");
    if(n !=0) {
        document.getElementById("titleLogo").src = "app/" + logo;
    } else {
        document.getElementById("titleLogo").src = logo;
    }
}

function setBrandingLogo(logo) {
    if(logo == null) {
        document.getElementById("apeLogo").style.display = "none";
    }

    var n = logo.indexOf("://");
    if(n == -1) {
        document.getElementById("apeLogo").src = "app/" + logo;
    } else {
        document.getElementById("apeLogo").src = logo;
    }
}

var opacitySetting = 0;

function fadeTitle(color) {

    opacitySetting = 0;

    document.getElementById("introFader").style.backgroundColor = color;
    document.getElementById("introFader").style.opacity = 0;
    document.getElementById("introFader").style.display = "block";


    doFade();
}

function doFade() {

    opacitySetting += .05;

    document.getElementById("introFader").style.opacity = opacitySetting;

    if(opacitySetting >= 1) {
        if(appFrame.contentWindow.webAppCore.titleFadeCallback != null) {
            appFrame.contentWindow.webAppCore.titleFadeCallback();
        }
    } else {
        setTimeout(doFade,50);
    }
}

function addTitleOption(name,tag,color,selectedColor) {
    var titleOption = document.createElement("p");
    titleOption.className = "introScreenOption";
    titleOption.innerHTML = name;
    titleOption.tagReference = tag;

    if(color && color != null) {
        titleOption.style.color = color;
        titleOption.standardColor = color;
    } else {
        titleOption.standardColor = "#ffffff";
    }

    if(selectedColor && selectedColor != null) {
        titleOption.selectedColor = selectedColor;
    } else {
        titleOption.selectedColor = "#FFB74D";
    }

    titleOption.addEventListener("click",function() {
        appFrame.contentWindow.webAppCore.titleOptionSelected(tag);
    },false);

    document.getElementById("introScreenOptions").appendChild(titleOption);
}

function clearTitleOptions() {
    document.getElementById("introScreenOptions").innerHTML = "";
}

function dataRequest(url,callback,data,post) {
    var xmlhttp = new XMLHttpRequest();
    xmlhttp.callbackFunction = callback;

    var m = "GET";

    if(post) {
        m = "POST";
    }

    var params = "";

    if(data && data != null) {
        if(Array.isArray(data)) {
            for(var i = 0;i < data.length; i++) {
                if(i > 0) {
                    params = params + "&";
                }

                if(post) {
                    params = params + data[i].key + "=" + data[i].value;
                } else {
                    params = params + data[i].key + "=" + encodeURIComponent(data[i].value);
                }
            }
        } else {
            if (typeof data === 'string' || data instanceof String) {
                params = data;
            } else {
                if(post) {
                    params = data.key + "=" + data.value;
                } else {
                    params = data.key + "=" + encodeURIComponent(data.value);
                }
            }
        }
    }

    xmlhttp.onreadystatechange=function() {
        if (xmlhttp.readyState==4 && xmlhttp.status==200) {
            this.callbackFunction(xmlhttp.responseText);
        }
    }

    if(!post && params.length > 0) {
        url = url + "?" + params;
    }

    xmlhttp.open(m,url,true);
    xmlhttp.setRequestHeader("Content-type", "application/x-www-form-urlencoded");

    if(post) {
        console.log("sending post");
        xmlhttp.send(params);
    } else {
        console.log("sending get");
        xmlhttp.send();
    }
}

function postForResult(url,data) {
    dataRequest(url,wacResult,data,true);
}

function getForResult(url,data) {
    dataRequest(url,wacResult,data,false);
}

function wacResult(data) {
    appFrame.contentWindow.webAppCore.resultFunction(data);
}

function getIsPremium() {
    return Android.getIsPremium();
}

function titlescreenGamepadDown(key) {

    if(key == "up" || key == "down") {
        var titleOptions = document.getElementById("introScreenOptions").getElementsByClassName("introScreenOption");

        if(key == "down") {
            selectedTitleScreenOption++;

            if(selectedTitleScreenOption >= titleOptions.length) {
                selectedTitleScreenOption = 0;
            }
        } else {
            selectedTitleScreenOption--;

            if(selectedTitleScreenOption < 0) {
                selectedTitleScreenOption = titleOptions.length - 1;
            }
        }

        if(titleOptions.length == 1) {
            selectedTitleScreenOption = 0;
        }

        setTitleOptionSelection();

        listScrollTimeout = setTimeout(redoPendingGamepadScroll,200);
    }

    if((key == "a" || key == "start") && selectedTitleScreenOption != -1) {
        document.getElementById("introScreenOptions").getElementsByClassName("introScreenOption")[selectedTitleScreenOption].click();
    }

    if(key == "left" || key == "select" || key == "lb" || key == "lt") {
        toggleMenu();
    }

    if((key == "rb" || key == "rt" || key == "rup") && window.useToolbar) {
        toggleToolbar();
    }
}

var pendingGamepadDirectionKey = null;
var pendingGamepadId = null;

function redoPendingGamepadScroll() {
    gamepadDown(pendingGamepadDirectionKey,pendingGamepadId);
}

function killScrollTimeout() {
    if(listScrollTimeout && listScrollTimeout != null) {
        clearTimeout(listScrollTimeout);
        listScrollTimeout = null;
    }

    pendingGamepadId = null;
}

function onFacebookStatusChange(status) {

    if(appFrame == null) {
        return;
    }

    if(appFrame.contentWindow.webAppCore) {
        if(appFrame.contentWindow.webAppCore.onFacebookStatusChanged) {
            appFrame.contentWindow.webAppCore.onFacebookStatusChanged(status);
        }
    }
}

function setTitleSubtitle(text) {

    if (text.trim().length == 0) {
        document.getElementById("titleSubtitle").style.display = "none";
        return;
    }

    document.getElementById("titleSubtitle").style.display = "block";
    document.getElementById("titleSubtitle").innerHTML = text;
}

function onToastActionPressed() {
    if(appFrame.contentWindow.webAppCore) {
        appFrame.contentWindow.webAppCore.toastBack();
    }
}

function notifyPlatformNavigation(url)  {
    startReportingShake = false;
    Android.setSensorActivation(false,0);

    sizeElements();
}

function wacImageBack(data) {
    appFrame.contentWindow.webAppCore.picImportCallbackHook(data);
}

function interfaceBackPress() {
    sizeElements();

    if(currentOpenDialog) {
        if(!currentOpenDialog.hidden) {
            currentOpenDialog.hide();
            return;
        }
    }

    if(activeDialog != null) {
        activeDialog.hide();
        return;
    }

    if(appFrame != null && appFrame.contentWindow.webAppCore != null) {
        var backFunction = appFrame.contentWindow.webAppCore.onBack();
        Android.log("onBack return: " + backFunction);

        if(backFunction != "0"){
            Android.log("back handled");
            return;
        }
    }

    Android.log("no back function set");

    if(appFrame != null && currentFrameURL != window.appUrl && appFrame.contentWindow.webAppCore != null) {
        Android.log("going back");
        appFrame.contentWindow.webAppCore.goBack();
        sizeElements();
        return;
    }

    Android.log("cant navigate back");
    Android.doFinish();
}

function juiceUpIAP() {
    // hrmm
}

function goPremium() {
    Android.goPremium();
}

// still needed for script.js, may change in future
function storeVar(key,value) {
    Android.saveVar(key,value.toString());
}

// still needed for script.js, may change in future
function getVar(key) {
    return Android.returnVar(key);
}

function toggleShareBox(value) {
    textToShare = value.replace("#appurl",window.shareURL);

    Android.nativeShare(textToShare);
}

function vibrate(duration) {
    Android.vibratePhone(duration);
}

function endVibrate() {
    Android.stopVibration();
}

function setTheme(color) {
    lastTheme = color;
    Android.setTheme(color);
}

function processScreenshot(canvas) {
    var dataURL = canvas.toDataURL();

    Android.saveScreenFromData(dataURL,false);

    showToast("Picture saved to Photos!");
}

function shareImage(element) {
    var blob;

    if (element) {
        var elementType = element.tagName.toLowerCase();

        if (elementType == "img") {
            var cv = document.createElement("canvas");
            cv.width = element.width;
            cv.height = element.height;
            var ct = cv.getContext("2d");
            ct.drawImage(element, 0, 0,element.width,element.height);
            blob = cv.toDataURL();
            Android.saveScreenFromData(blob,true);
            return;
        }

        if (elementType == "canvas") {
            blob = element.toDataURL();
            Android.saveScreenFromData(blob,true);
            return;
        }
    } else {
        element = appFrame.contentWindow.document.body;
    }

    html2canvas(element, {
        onrendered: function (canvas) {
            var blob = canvas.toDataURL();
            Android.saveScreenFromData(blob,true);
        }
    });
}

function saveFile(filename,filecontent) {

    if(typeof filecontent === "string") {
        Android.saveFile(filename,filecontent);
        return;
    }

    if(typeof filecontent === "number") {
        Android.saveFile(filename,filecontent.toString());
        return;
    }


    var uInt8Array = new Uint8Array(filecontent);
    Android.saveFileBytes(filename,uInt8Array);
}

function getFiles() {
    var listing = Android.getFileListing();

    var retArray = new Array();
    var rawArray = listing.split("|");

    for(var i = 0; i < rawArray.length; i++) {
        if(rawArray[i].length > 0) {
            retArray[i] = rawArray[i];
        }
    }

    appFrame.contentWindow.webAppCore.listingLoaded(retArray);
}

function deleteFile(filename) {
    Android.deleteFile(filename);
}

function processAnalytics() {
    // hrmm
}

function loadURL(url) {
    Android.navigateExternal(url);
}

function doTTS(text) {
    Android.doTTS(text);
}

function showToast(message,icon,actionButton) {

    var icn = null;
    var act = null;

    if(icon && icon != null) {
        icn = icon;
    }

    if(actionButton && actionButton != null) {
        act = actionButton.title;
    }

    Android.showToast(message,icn,act);
}

function submitHighScore(modeid,score) {
    Android.submitHighScore(modeid.toString(),score.toString());
}

function prepareAchievementData() {
    // handled by external wrapper
}

function reportAchievement(trophy,increment) {
    Android.submitAchievement(trophy.toString(),increment);
}

function showAchievements() {
    Android.viewAchievements();
}

function setHighScoresUsername() {

}

function getHighScoresUsername() {

}

function getDailyTopScore(mode) {
    // hrmm
    /*
    if(scoreServerId === 'undefined') {
        return;
    }

    var scoreData = new Array();
    scoreData[scoreData.length] = {key:"g",value:scoreServerId};
    scoreData[scoreData.length] = {key:"m",value:mode};


    dataRequest("http://www.ape-apps.com/highscores/app_resource.php?f=2",processTopScoreResult,scoreData,true);
    */
}

function processTopScoreResult(data) {
    if (data.indexOf("###") == -1) {
        return;
    }

    var resultReadout = "Today's top score: " + data.replace("###"," - ");

    appFrame.contentWindow.webAppCore.setDailyReadout(resultReadout);
}

function toggleScores(mode) {
    Android.viewHighScores(mode.toString());
}

function clearPlatformSplash() {
    if (typeof Android.clearSplash == 'function') {
        Android.clearSplash();
    }
}

// Web App Core 3.1 Hooks
function importLocalFile(type) {
    Android.importFile(type);
}

function wacListCallback(tag) {
    appFrame.contentWindow.webAppCore.nativeListOptionSelected(tag);
}

function notifyImportReady() {
    var name = Android.getImportName();
    var content = Android.getImportContent();

    appFrame.contentWindow.webAppCore.fileImportComplete(name,content);
}

function importComplete(name,contents) {
    appFrame.contentWindow.webAppCore.fileImportComplete(name,contents);
}

function exportTextFileAsDialog(filename,content,isXML) {

    if(typeof content === "string") {
        Android.exportFile(filename,content);
        return;
    }

    var uInt8Array = new Uint8Array(content);
    Android.exportFileBytes(filename,uInt8Array);
}

function sendFile(filename) {
    Android.sendFile(filename);
}

function copyToClipboard(text) {
    Android.copyText(text);
}

function getInterface() {
    return "android";
}

function isTV() {
    return Android.getIsTvDevice();
}

function printContent(content) {
    Android.printDocument(content);
}

function changeWorkingDirectory() {
    Android.moveWorkspace();
}

function workspaceChangeCallback(result) {
    appFrame.contentWindow.webAppCore.workspaceChanged(result);
}

var packSaveTimer;
var tmpPackageName;
var tmpPackageContent;

function updateVarPackage(pkg) {

    if(packSaveTimer) {
        clearTimeout(packSaveTimer);
    }

    tmpPackageContent = JSON.stringify(pkg);
    tmpPackageName = window.webAppId + "_var_package.wvp";

    packSaveTimer = setTimeout(function() {
        packSaveTimer = null;
        Android.saveFile(tmpPackageName,tmpPackageContent);
    },500);

}

function buildVarPackage() {

    var fileName = window.webAppId + "_var_package.wvp";

    var content = Android.openFile(fileName);

    if(content && content.trim().length > 0) {
        var obj = JSON.parse(content);
        appFrame.contentWindow.webAppCore.recieveVariablePackage(obj);
    } else {
        appFrame.contentWindow.webAppCore.recieveVariablePackage(new Object());
    }
}

var startReportingShake = false;

function setShakeSensitivity(sensitivity) {
    startReportingShake = true;
    Android.setSensorActivation(true,sensitivity);
}

function sakeEventCallback() {
    if(startReportingShake) {
        if(appFrame.contentWindow.webAppCore.shakeCallback) {
            appFrame.contentWindow.webAppCore.shakeCallback();
        }
    }
}

function checkForLaunchFile() {
    var hasLaunchFile = Android.doesLaunchFileExist();

    if(appFrame.contentWindow.webAppCore.launchCheckCallback) {
        appFrame.contentWindow.webAppCore.launchCheckCallback(hasLaunchFile);
    } else {
        console.log("no launch callback!");
    }
}

function getLaunchFileName() {
    var launchFileName = Android.getLaunchFileName();

    if(appFrame.contentWindow.webAppCore.launchNameCallback) {
        appFrame.contentWindow.webAppCore.launchNameCallback(launchFileName);
    }
}

function getLaunchFileContent() {
    var launchFileContent = Android.openLaunchFile();

    if(appFrame.contentWindow.webAppCore.launchContentCallback) {
        appFrame.contentWindow.webAppCore.launchContentCallback(launchFileContent);
    }
}

function saveLaunchFile(content) {
    Android.saveLaunchFile(content);
}

function onGeneralError(e) {
    var appVersion = Android.getAppVersion();
    var scrpt = e.filename.substr(e.filename.lastIndexOf('/') + 1);

    Android.trackEvent('Android Web Frame Error ' + appVersion,scrpt + ":" + e.lineno,e.message, false);
}

function onFrameError(e) {
    var appVersion = Android.getAppVersion();
    var scrpt = e.filename.substr(e.filename.lastIndexOf('/') + 1);

    Android.trackEvent('Android Webapp Error ' + appVersion,scrpt + ":" + e.lineno,e.message, false);
}

var audioArray;
var loopingAudio;
var audioType = "ogg";

function playAudio(soundname) {
    var soundFileURL = "app/audio/" + audioType + "/" + soundname + "." + audioType;
    playAudioChannel(soundFileURL,0);
}

function playAudioChannel(soundName,channel) {

    if(!audioArray[channel]) {
        audioArray[channel] = [];
    }

    var arrayToUse = audioArray[channel];

    var audioObject;

    for(var i = 0; i < arrayToUse.length; i++) {
        if(arrayToUse[i].name == soundName) {
            audioObject = arrayToUse[i];
        }
    }

    if(audioObject) {
        if(audioObject.audObj.currentTime > 0) {
            playAudioChannel(soundName,(channel + 1));
            return;
        }
        audioObject.audObj.play();
    } else {
        var myAudio = document.createElement('audio');
        myAudio.src = soundName;
        myAudio.addEventListener("ended",function(e) {
            this.currentTime = 0;
        },false);
        myAudio.play();

        audioObject = { name:soundName,audObj:myAudio };

        arrayToUse.push(audioObject);
    }
}

function loopAudio(soundname) {
    var soundFileURL = "app/audio/" + audioType + "/" + soundname + "." + audioType;

    loopingAudio.src = soundFileURL;
    loopingAudio.loop = true;
    loopingAudio.play();
}

function killLoop() {
    loopingAudio.pause();
    loopingAudio.src = "";
    try {
        loopingAudio.currentTime = 0;
    } catch(err) {
        // whatevs
        ga('send', 'exception', {
            'exDescription': err.message,
            'exFatal': false
        });
    }
}

function wacSoundInstanceDone(soundName) {
    if(appFrame.contentWindow.webAppCore.audioInstanceCallbackListener) {
        appFrame.contentWindow.webAppCore.audioInstanceCallbackListener(soundName);
    }
}

function wacRewardDone(status) {
    if(appFrame.contentWindow.webAppCore.rewardedVideoDone) {
        appFrame.contentWindow.webAppCore.rewardedVideoDone(status);
    }
}

function wacNativeGetBack(id,data) {
    if(appFrame.contentWindow.webAppCore.nativeGetBack) {
        appFrame.contentWindow.webAppCore.nativeGetBack(id,data);
    }
}